# (C) 2020, Schlumberger. Refer to LICENSE. 
'''
 agnostic : to support code that is CPU and GPU compatible via cupy
            Unfortunately cupy cannot be installed on a machine that
            does not have GPU so we also have cupy_facade.py and do
            not include cupy in the requirements.txt

 When considering developing GPU algorithms refer to:
 https://docs-cupy.chainer.org/en/stable/reference/comparison.html

 The general policy is to have numpy/ndimage/scipy algorithms that are
 GPU-enabled by cupy. In this sense, GPU is secondary to CPU.
'''
HAS_GPU=False
try:
    import cupy as cp
    HAS_GPU=True
except:
    import distpy.calc.cupy_facade as cp

import numpy
import scipy
from scipy import ndimage


class agnostic(object):

    # Don't use this in distpy projects...
    #def get_array_module(self,a):
    #    return cp.get_array_module(a)

    # This is used in place of cp.get_array_module(a)
    # because in the future we might also want get_scipy_ndimage (see below)
    def get_numpy(self,a):        
        return cp.get_array_module(a)


    # Currently none of the extra_numpy algorithms using ndimage
    # are GPU compatible.
    def get_scipy_ndimage(self,a):
        global HAS_GPU
        if HAS_GPU==True:
            return cp.get_array_module(a)
        else:
            return ndimage

    def asarray(self,a):
        return cp.asarray(a)

    def asnumpy(self,a):
        return cp.asnumpy(a)



from __future__ import division, absolute_import, print_function

__all__ = ['parallel_convert_las2fbe',
           'parallel_ingest_csv',
           'parallel_ingest_h5',
           'parallel_ingest_sgy',
           'parallel_ingest_witsml']






# (C) 2020, Schlumberger. Refer to LICENSE

import numpy
import datetime
import scipy.signal
import os
import distpy.io_help.io_helpers as io_helpers
import distpy.io_help.directory_services as directory_services
import distpy.calc.pub_command_set as pub_command_set

import distpy.calc.extra_numpy as extra_numpy
import distpy.calc.processing_commands as processing_commands


def build_command_list(commandJson, data=None, dirout="scratch",
                       datedir="none",datestring="none",nx=100,prf=10000,xaxis=None,taxis=None,nt=100,
                       extended_list=[],inline_plots=0):
    if xaxis is None:
        xaxis=numpy.arange(nx)
    if taxis is None:
        taxis=numpy.arange(nt)
    # initialize the command list that will eventually be executed.
    # This first one is anomalous - need a good way to initiate the top of the tree...
    # possibly this is a root...
    command_list=[]
    # { "name" : "data",           "uid" :  0},
    command_list.append(pub_command_set.DataLoadCommand(data,{}))
    
    # command_list is a list of dictionaries
    for command in commandJson['command_list']:
        # we add the basics of nx and prf, which belong to the data
        dir_suffix = command.get('directory_out','NONE')
        if not dir_suffix=='NONE':
            dirval = os.path.join(dirout,dir_suffix)
            if not os.path.exists(dirval):
                os.makedirs(dirval)
            command['directory_out']=dirval
        dir_suffix = command.get('directory_in','NONE')
        if not dir_suffix=='NONE':
            dirval = os.path.join(dirout,dir_suffix)
            if not os.path.exists(dirval):
                os.makedirs(dirval)
            command['directory_in']=dirval
        command['date_dir']=datedir
        command['datestring']=datestring
        command['nx']=nx
        command['prf']=prf
        command['xaxis']=xaxis
        command['taxis']=taxis
        command['nt']=nt
        command['inline_plots']=inline_plots
        # prf/nt is the rescale factor
        command['f_rescale']=float(nt)/float(prf)
        index = command.get('band00',-1)
        if index>=0:
            command['command']=command_list[index]
        #s(len(command_list))
        command_list.append(processing_commands.CommandFactory(command_list,command,extended_list=extended_list))
    return command_list

def docs(command_list,commandJson):
    ltxJson = []
    ltxJson = io_helpers.latexJson({"command_list" : commandJson['command_list']},ltxJson)

    lines =[]
    lines = io_helpers.latexTop(lines)
    lines = io_helpers.command2latex(command_list,lines, commandJson.get('name','Command Set'), commandJson.get('description','Commands used'))

    for line in ltxJson:
        lines.append(line)
    lines = io_helpers.latexPng(lines)
    lines = io_helpers.latexTail(lines)

    # dot graphviz graph
    graphs = io_helpers.dot_graph(commandJson['command_list'])
    return lines, graphs
'''
 strainrate2summary : process one file of strainrate data using the provided
                      command tree in commandJSON.

                      This executes in one thread.
'''
def strainrate2summary(filename, xaxis, prf, dirout, commandJson, extended_list):
    # try to make a datestamp that WITSML could use...
    tokens = filename.split(os.sep)
    # the final token without its .npy is the unix timestamp
    unixtime = int(tokens[-1][:-4])
    datestring = datetime.datetime.utcfromtimestamp(unixtime).strftime("%Y-%m-%dT%H:%M:%S+00:00")
    datedir = str(unixtime)



    # Configure the hardware
    boxsize = commandJson.get('BOXSIZE', 500)
    extra_numpy.set_boxsize(boxsize)
    data = numpy.load(filename)
    #print(data.shape)
    nx = data.shape[0]
    nt = data.shape[1]

    # Is this a request to document a workflow?
    isDocs = commandJson.get('document',0)

    # initialize the command list that will eventually be executed.
    # This first one is anomalous - need a good way to initiate the top of the tree...
    # possibly this is a root...
    taxis = commandJson.get('taxis',None)
    # A convenience for Jupyter Notebooks - when developing workflows we sometimes want to inline plot views
    inline_plots = commandJson.get('inline_plots',0)

    command_list=build_command_list(commandJson, data=data,dirout=dirout,
                                    datedir=datedir,datestring=datestring,
                                    nx=nx,prf=prf,xaxis=xaxis,taxis=taxis,
                                    nt=nt,extended_list=extended_list,inline_plots=inline_plots)

    # if documenting - that is one path...execution is the other
    if isDocs>0:
        lines, graphs = docs(command_list, commandJson)
        with open(os.path.join(dirout,'config.tex'),'w') as f:
            for line in lines:
                f.write(line+'\n')
                
        # dot graphviz graph
        with open(os.path.join(dirout,'config.gv'),'w') as f:
            for line in graphs:
                f.write(line+'\n')
        
    else:
        # if postconditions are all already met, we can skip this workflow...
        post_cond = True
        for command in command_list:
            for filename in command.postcond():
                if not directory_services.exists(filename):
                    post_cond = False

        if post_cond==True:
            print(filename, ' post-conditions are all satisfied, skipping processing.')
        else:
            for command in command_list:
                command.execute()
    return None, None



# (C) 2020, Schlumberger. Refer to LICENSE
'''
 plt_command_set.py

 Collecting all the plot commands in one place

 The BasicCommand can be found in distpy.calc.pub_command_set
 All commands in here must inherit from BasicCommand
 Most commands in here should inherit from BasicPlotCommand, which
     extends BasicCommand in a way that makes matplotlib plots easy
     to generate with consistent annotation
'''
import numpy
import os
import matplotlib.pyplot as plt
from numba import vectorize, float32

import distpy.calc.extra_numpy as extra_numpy
import distpy.calc.extra_pyplot as extra_pyplot
import distpy.io_help.io_helpers as io_helpers
import distpy.io_help.directory_services as directory_services
from distpy.calc.pub_command_set import BasicCommand
from distpy.calc.pub_command_set import universal_arglist

'''
 ThumbnailCommand : turn the current 2D matrix into an image, which can be aliased.
                    This is commonly called in the middle of a signal processing flow to
                    generate interim result plots.
'''
class ThumbnailCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._prevstack = jsonArgs.get('commands',[None])
        self._dirname = jsonArgs.get('directory_out','NONE')
        self._fname = jsonArgs.get('date_dir','NONE')
        if "xaxis" in jsonArgs:
            self._xaxis = [jsonArgs['xaxis'][0],jsonArgs['xaxis'][-1]]
            self._taxis = [0,jsonArgs['nt']/jsonArgs['prf']]
        else:
            self._xaxis = [0,1]
            self._taxis = [0,1]
        if "taxis" in jsonArgs:
            self._taxis = [jsonArgs['taxis'][0],jsonArgs['taxis'][-1]]
        self._format = jsonArgs.get('format','png')
        self._std_val_mult = jsonArgs.get('clip_level',1.0)
        self._jsonArgs = jsonArgs
        

    def docs(self):
        docs={}
        docs['one_liner']="Create a thumbnail image of the current 2D processed data in the specified format."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['directory_out','format','clip_level'] }
        return docs

    def postcond(self):
        return [directory_services.path_join(self._dirname,self._fname)]

    def execute(self):
        super().execute()
        datalist = None
        if len(self._prevstack)>0:
            if self._prevstack[0] is not None:
                datalist = []
                for dataset in self._prevstack:
                    datalist.append(dataset.result())

        dirname = self._dirname
        if not dirname=='NONE':
            io_helpers.thumbnail_plot(dirname,self._fname,self._previous.result(),
                                      xscale=self._xaxis,tscale=self._taxis,data2=datalist,
                                      plt_format=self._format, std_val_mult=self._std_val_mult, jsonArgs=self._jsonArgs)

'''
 BasicPlotCommand - in particular extends BasicCommand with annotations() and write_plot()
                    to make it easy to generate consistent plot annotations (e.g. depth and time ranges overlain on the pictures)
'''
class BasicPlotCommand(BasicCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)
        self._indir = jsonArgs.get('directory_in','NONE')
        self._dirname = jsonArgs.get('directory_out','NONE')
        self._out_plot = jsonArgs.get('out_plot','NONE')

        # For all plots...
        self._depth_display_unit = jsonArgs['depth_display_unit']
        self._start_of_fibre = jsonArgs['start_of_fibre']
        self._end_of_fibre = jsonArgs['end_of_fibre']
        self._figure_size = jsonArgs['figure_size']
        self._dots_per_inch = jsonArgs['dpi']
        self._event_list = jsonArgs['label_list']
        ## clusters and stages - file location
        self._well_segs = jsonArgs['well_segments']
        self._depth_ref = jsonArgs['depth_ref']
        self._time_ref = jsonArgs['time_ref']
        self._nt = self._time_ref.shape[0]
        self._nx = self._depth_ref.shape[0]

        # For a specific plot...
        self._plot_list = jsonArgs['plot_list']
        self._plot_type = jsonArgs['plot_type']
        self._out_plot = jsonArgs['out_plot']

    def annotations(self, plt, title_text, t_interval):
        # how are plots and figures passed around?
        # here we are assuming that plt = matplotlib.pyplot can handle it
        # by assumption...not a good strategy
        if (t_interval<1):
            t_interval=1
        time_list = extra_pyplot.time_stamps(self._time_ref)
        nt = self._nt
        plt.title(title_text, fontsize=10)
        plt.xticks(numpy.arange(0, (nt-t_interval), t_interval), time_list[0:nt:t_interval],rotation='vertical',fontsize=7)      ## plot x ticks and labels
        plt.xlim(0,nt)
        plt.gca().axes.invert_xaxis()
        plt.axvline(0,color='k',linewidth=5)
        plt.ylim(numpy.max(self._depth_ref),numpy.min(self._depth_ref))

        extra_pyplot.add_user_events(self._event_list, plt, self._time_ref)
        extra_pyplot.add_fibre_ends(self._start_of_fibre, self._end_of_fibre, plt)
        extra_pyplot.add_user_zones(self._well_segs, plt, self._depth_ref, 'r')

    def write_plot(self,plt):
        super().execute()
        dirname = self._dirname
        if not dirname=='NONE':
            io_helpers.write_plot(dirname,self._out_plot, plt)

    def postcond(self):
        return [directory_services.path_join(self._dirname,self._out_plot)]

    def execute(self):
        super.execute()

'''
 WellLogPlotCommand : plot the data as a well-log, the basic "waterfall"
'''
class WellLogPlotCommand(BasicPlotCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)

    def execute(self):
        plot_list = self._plot_list
        nt = self._nt
        depth_ref = self._depth_ref
        n_bands = len(plot_list)
        # Equal plot widths for multiple panels (e.g. all the bands on one plot)
        h_len = 0.966/(n_bands+2) 
        for a in range(n_bands):
            local_plot_list = plot_list[a]
            fname = os.path.join(self._indir,local_plot_list['data'])
            traces = numpy.load(fname)
            
            title_text = local_plot_list['title_text']
            cscale = extra_pyplot.supported_colorscales(local_plot_list.get('colormap','magma'))
            
            vscale     = numpy.max(extra_numpy.despike(traces,3))               ##despike and get max value for scaling 2 *sdev
            vscalemax = vscale*10
            vscalemin=0
            t_interval = int(nt/8)
            # data is read, so now create the plot
            # PLOT...
            # NOTE: fontsize is hard coded!!! 
            plt.axes([0.15 + (a * h_len), 0.1, h_len, .8])
            plt.imshow(traces, extent=[nt,0, numpy.max(depth_ref), numpy.min(depth_ref)], cmap=cscale, aspect='auto', vmin=vscalemin,vmax=vscalemax)
            # standard finishing...
            self.annotations(plt, title_text, t_interval)
        self.write_plot(plt)

'''
  StackPlotCommand : multiple datasets (e.g. FBE bands) overlain at different transparency levels - which is a useful false color plotting technique
'''
class StackPlotCommand(BasicPlotCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)

    def execute(self):
        plot_list = self._plot_list
        nt = self._nt
        depth_ref = self._depth_ref

        n_bands = len(plot_list)
        # Equal plot widths for multiple panels (e.g. all the bands on one plot)
        title_text = 'Stack of '
        t_interval = int(nt/8)
        for a in range(n_bands):
            local_plot_list = plot_list[a]
            fname = os.path.join(self._indir,local_plot_list['data'])
            traces = numpy.load(fname)
            
            title_text = title_text + local_plot_list['title_text'] + ' '
            cscale = local_plot_list['colormap']
            alpha_val = float(local_plot_list['alpha'])
                        
            vscale     = numpy.max(extra_numpy.despike(traces,3))               ##despike and get max value for scaling 2 *sdev
            vscalemax = vscale*10
            vscalemin=0
            # data is read, so now create the plot
            # PLOT...
            plt.imshow(traces, extent=[nt,0, numpy.max(depth_ref), numpy.min(depth_ref)], cmap=cscale, alpha=alpha_val, aspect='auto', vmin=vscalemin,vmax=vscalemax)
        self.annotations(plt, title_text, t_interval)
        self.write_plot(plt)

'''
 RGBPlotCommand : plot 3 results (e.g. FBE bands) as the R, G and B intensities of an image, another false-color plotting technique
'''
class RGBPlotCommand(BasicPlotCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)

    def execute(self):
        plot_list = self._plot_list
        nt = self._nt
        nx = self._nx
        depth_ref = self._depth_ref
        t_interval = int(nt/8)
        red = plot_list[0]
        green = plot_list[1]
        blue = plot_list[2]
        title_text = 'R = ' + red['title_text'] + ', G = ' + green['title_text'] + ', B = ' + blue['title_text']
        
        img = numpy.zeros((nx, nt, 3), dtype=float)
        fname = os.path.join(self._indir,red['data'])
        img[:,:,0] = numpy.load(fname)
        fname = os.path.join(self._indir,green['data'])
        img[:,:,1] = numpy.load(fname)
        fname = os.path.join(self._indir,blue['data'])
        img[:,:,2] = numpy.load(fname)
        
        img[:,:,0]   = extra_numpy.intensity(img[:,:,0]) 
        img[:,:,1]   = extra_numpy.intensity(img[:,:,1])
        img[:,:,2]   = extra_numpy.intensity(img[:,:,2])
        
        # support inverted (so that (0,0,0) maps to white instead of black)
        # this is an optional argument so we are using the get() method instead of []
        if (red.get('inverted','no')=='yes'):
            img[:,:,0] = 1 - img[:,:,0]
        if (green.get('inverted','no')=='yes'):
            img[:,:,1] = 1 - img[:,:,1]
        if (blue.get('inverted','no')=='yes'):
            img[:,:,2] = 1 - img[:,:,2]
        
        img = img*255
                    
        plt.imshow((img.astype(numpy.uint8)), extent=[nt,0, numpy.max(depth_ref), numpy.min(depth_ref)], aspect='auto')
        self.annotations(plt,title_text, t_interval)
        self.write_plot(plt)


'''
  KnownCommands : the list of commands in this file. If a command is not listed here, it means that it is not accessible via distpy
'''
def KnownCommands(knownList):
    knownList['thumbnail']      = ThumbnailCommand
    knownList['well_log']       = WellLogPlotCommand
    knownList['stack']          = StackPlotCommand
    knownList['rgb']            = RGBPlotCommand
    return knownList



# (C) 2020, Schlumberger. Refer to LICENSE
'''
 sgy : A module for reading SEG-Y files in 1-second chunks.
       Source document:
       https://seg.org/Portals/0/SEG/News%20and%20Resources/Technical%20Standards/seg_y_rev2_0-mar2017.pdf

       The headers are recorded into JSON format, which makes this reader slow. It is not intended to be high performance
       or to replace other useful SEGY readers such as obspy and segyio. The intention is to provide a useful tool for
       researchers to rapidly prototype, so complete data ingestion with header preservation was the priority

       IBM Float (data_type 1) is supported as we have seen vendors using it in DAS segy files

       However the following types are not current supported
       4  : 4-byte fixed-point with gain (obsolete)
       7  : 3-byte twos complement integer
       15 : 3-byte unsigned integer

       The ascii header is assumed to be cp500 format. If the ascii header looks like garbage try utf-8
       
       An attempt is made to correct the most common errors found in current DAS files that are in the wild.
       In particular some vendors writers fail to write the number of samples in a trace in the correct place
       when that number is larger than the SEG-Y rev.1 maximum. A fix is made to put match the number of samples
       with the correct header location for SEG-Y rev.2, which can handle very long traces.
       The aim is that, in the future, a SEGY writer could be constructed based on the JSON headers and npy files, and
       that recovered file would then be a well-formed SEG-Y rev.2
'''
import codecs
import os
import struct
import json
# This is the only numba dependency - we use it to vectorize the ibm2ieee
# https://github.com/numba/numba/blob/master/LICENSE

import numpy
import datetime
from numba import vectorize, float32, uint64, float64
import distpy.io_help.directory_services as directory_services

# If the SEGY is using the old-school IBM floats...
# this snippet is from https://stackoverflow.com/questions/7125890/python-unpack-ibm-32-bit-float-point
@vectorize([float64(uint64)])
def ibm2ieee(ibm):
    """
    Converts an IBM floating point number into IEEE format.
    :param: ibm - 32 bit unsigned integer: unpack('>L', f.read(4))
    """
    if ibm == 0:
        return 0.0
    sign = ibm >> 31 & 0x01
    sign = numpy.double(sign)
    exponent = ibm >> 24 & 0x7f
    exponent = numpy.double(exponent)-64.0;

    mantissa = (ibm & 0x00ffffff)
    mantissa = numpy.double(mantissa)
    mantissa = mantissa / numpy.double(numpy.power(2,24))

    return (1 - 2 * sign) * mantissa * numpy.power(16, exponent)

'''
 file_header : Read the main header.
'''
def file_header(fullHeader):
    # Useful constants and format codes
    formats = {}
    formats['ENDIAN'] = 'big'
    formats['DOUBLE'] = '>d'
    formats['UNSIGNED_LONG_LONG'] = '>Q'
    formats['UNSIGNED_CHAR'] = 'B'
    formats['UNSIGNED_SHORT'] = '>H'
    formats['UNSIGNED_INT'] = '>I'
    formats['INT'] = '>i'
    formats['SHORT'] = '>h'
    formats['EBCDIC'] = 'cp500'


    # Conversion to JSON - use python dictionaries
    header = {}
    # REV 2 - the endian test integer...check this first.
    header['endian_test_constant'] = struct.unpack(formats['UNSIGNED_INT'],fullHeader[3297:3301])[0]
    
    # Update the format codes now we konw the format
    if (header['endian_test_constant']==67305985):
        formats['ENDIAN']='little'
        formats['DOUBLE'] = '<d'
        formats['UNSIGNED_LONG_LONG'] = '<Q'
        formats['UNSIGNED_CHAR'] = 'B'
        formats['UNSIGNED_SHORT'] = '<H'
        formats['UNSIGNED_INT'] = '<I'
        formats['INT'] = '<i'
        formats['SHORT'] = '<h'

    #SEGY 3200 byte header
    #40 lines, 80 bytes each
    istep = 80
    iend = istep
    istart = 0
    asciiHeader = ""
    for item in range(40):
        byte80Part=fullHeader[istart:iend]
        istart+=istep
        iend+=istep
        asciiHeader += byte80Part.decode(formats['EBCDIC'])
        asciiHeader += "\n"
    header['ascii_header']=asciiHeader
    #print(asciiHeader)
    # https://seg.org/Portals/0/SEG/News%20and%20Resources/Technical%20Standards/seg_y_rev2_0-mar2017.pdf
    # table 2
    # Local variables for convenience
    ENDIAN = formats['ENDIAN']
    DOUBLE = formats['DOUBLE']
    UNSIGNED_LONG_LONG = formats['UNSIGNED_LONG_LONG']
    UNSIGNED_CHAR = formats['UNSIGNED_CHAR']
    UNSIGNED_SHORT = formats['UNSIGNED_SHORT']
    UNSIGNED_INT = formats['UNSIGNED_INT']
    INT = formats['INT']
    SHORT = formats['SHORT']   
    header['job_id'] = struct.unpack(INT,fullHeader[3200:3204])[0]
    header['line_number'] = struct.unpack(INT,fullHeader[3204:3208])[0]
    header['reel_number'] = struct.unpack(INT,fullHeader[3208:3212])[0]
    header['number_of_data_traces_per_ensemble'] = struct.unpack(SHORT,fullHeader[3212:3214])[0]
    header['number_of_auxiliary_traces_per_ensemble'] = struct.unpack(UNSIGNED_SHORT,fullHeader[3214:3216])[0]
    header['sample_interval'] = struct.unpack(UNSIGNED_SHORT,fullHeader[3216:3218])[0]
    header['sample_interval_of_original_field_recording'] = struct.unpack(UNSIGNED_SHORT,fullHeader[3218:3220])[0]
    header['number_of_samples_per_data_trace'] = struct.unpack(UNSIGNED_SHORT,fullHeader[3220:3222])[0]
    header['number_of_samples_per_data_trace_for_original_field_recording'] = struct.unpack(UNSIGNED_SHORT,fullHeader[3222:3224])[0]
    header['data_sample_format'] = struct.unpack(SHORT,fullHeader[3224:3226])[0]
    header['ensemble_fold'] = struct.unpack(SHORT,fullHeader[3226:3228])[0]
    header['trace_sorting_code'] = struct.unpack(SHORT,fullHeader[3228:3230])[0]
    header['vertical_sum_code'] = struct.unpack(SHORT,fullHeader[3230:3232])[0]
    header['sweep_frequency_at_start'] = struct.unpack(SHORT,fullHeader[3232:3234])[0]
    header['sweep_frequency_at_end'] = struct.unpack(SHORT,fullHeader[3234:3236])[0]
    header['sweep_length'] = struct.unpack(SHORT,fullHeader[3236:3238])[0]
    header['sweep_type_code']= struct.unpack(SHORT,fullHeader[3238:3240])[0]
    header['trace_number_of_sweep_channel'] = struct.unpack(SHORT,fullHeader[3240:3242])[0]
    header['sweep_trace_taper_length_at_start'] = struct.unpack(SHORT,fullHeader[3242:3244])[0]
    header['sweep_trace_taper_length_at_end'] = struct.unpack(SHORT,fullHeader[3244:3246])[0]
    header['taper_type'] = struct.unpack(SHORT,fullHeader[3246:3248])[0]
    header['correlated_data_traces'] = struct.unpack(SHORT,fullHeader[3248:3250])[0]
    header['binary_gain_recovered'] = struct.unpack(SHORT,fullHeader[3250:3252])[0]
    header['amplitude_recovery_method'] = struct.unpack(SHORT,fullHeader[3252:3254])[0]
    header['measurement_system'] = struct.unpack(SHORT,fullHeader[3254:3256])[0]
    header['impulse_signal_polarity'] = struct.unpack(SHORT,fullHeader[3256:3258])[0]
    header['vibratory_polarity_code'] = struct.unpack(SHORT,fullHeader[3258:3260])[0]
    header['extended_traces_per_ensemble'] = struct.unpack(INT,fullHeader[3260:3264])[0]
    header['extended_auxiliary_traces_per_ensemble'] = struct.unpack(INT,fullHeader[3264:3268])[0]
    header['extended_samples_per_trace'] = struct.unpack(INT,fullHeader[3268:3272])[0]
    header['extended_sample_interval'] = struct.unpack(DOUBLE,fullHeader[3272:3280])[0]
    header['extended_sample_interval_of_original'] = struct.unpack(DOUBLE,fullHeader[3280:3288])[0]
    header['extended_samples_per_trace_in_original'] = struct.unpack(INT,fullHeader[3288:3292])[0]
    header['extended_ensemble_fold'] = struct.unpack(INT,fullHeader[3292:3296])[0]
    header['endian_test_constant'] = struct.unpack(INT,fullHeader[3296:3300])[0]
    # 3301-3500 - rev 2
    header['segy_major_rev'] = struct.unpack(UNSIGNED_CHAR,fullHeader[3500:3501])[0]
    header['segy_minor_rev'] = struct.unpack(UNSIGNED_CHAR,fullHeader[3501:3502])[0]
    header['fixed_length_trace_flag'] = struct.unpack(SHORT,fullHeader[3502:3504])[0]
    header['number_of_extended_textual_headers'] = struct.unpack(SHORT,fullHeader[3504:3506])[0]
    header['max_additional_240byte_headers'] = struct.unpack(INT,fullHeader[3506:3510])[0]
    header['time_basis_code'] = struct.unpack(UNSIGNED_SHORT,fullHeader[3510:3512])[0]
    header['number_of_traces_in_file'] = struct.unpack(UNSIGNED_LONG_LONG,fullHeader[3512:3520])[0]
    header['byte_offset_of_first_trace'] = struct.unpack(UNSIGNED_LONG_LONG,fullHeader[3520:3528])[0]
    header['number_of_trailer_stanzas'] = struct.unpack(INT,fullHeader[3528:3532])[0]
    #3533-3600 unassigned
    
    # Correct a problem from older DAS file sthat failed to write the SEGY header correctly
    if (header['number_of_samples_per_data_trace']==55537)&(header['max_additional_240byte_headers']>55537):
        print('FOUND BAD SEGY...trying to fix')
        header['extended_samples_per_trace'] = header['max_additional_240byte_headers']
        header['max_additional_240byte_headers'] = 0

    if header['extended_samples_per_trace']==0:
        header['extended_samples_per_trace']=header['number_of_samples_per_data_trace']
    # QC for the header - using json pretty printing
    #print (json.dumps(header, sort_keys=True, indent=4, separators=(',', ': ')))

    # Next there are the Extended Textual File Headers
    etfh={}
    ETFH_LENGTH=3200
    for a in range(header['number_of_extended_textual_headers']):
        etfh[a]=rawObj.read(ETFH_LENGTH)
    #print (json.dumps(etfh, sort_keys=True, indent=4, separators=(',', ': ')))

    # Another common error in DAS SEGY files...
    if header['number_of_traces_in_file']==0:
        if header['ensemble_fold']==0:
            # Special case - stacked VSP...
            if header['number_of_data_traces_per_ensemble']==0:
                print("The ensemble_fold was not specified")
            else:
                header['number_of_traces_in_file']=header['number_of_data_traces_per_ensemble']
        else:
          header['number_of_traces_in_file']=header['ensemble_fold']
        
    print(header)
    return header, formats

'''
 trace_header : read the header from a given trace into a dictionary, for later writing to JSON
'''
def trace_header(stdHeaderBytes, formats):
    ENDIAN = formats['ENDIAN']
    DOUBLE = formats['DOUBLE']
    UNSIGNED_LONG_LONG = formats['UNSIGNED_LONG_LONG']
    UNSIGNED_CHAR = formats['UNSIGNED_CHAR']
    UNSIGNED_SHORT = formats['UNSIGNED_SHORT']
    UNSIGNED_INT = formats['UNSIGNED_INT']
    INT = formats['INT']
    SHORT = formats['SHORT']            
    stdHeader = {}
    stdHeader['trace_sequence_number_within_line']=struct.unpack(INT,stdHeaderBytes[0:4])[0]
    stdHeader['trace_sequence_number_within_file']=struct.unpack(INT,stdHeaderBytes[4:8])[0]
    stdHeader['original_field_record_number']=struct.unpack(INT,stdHeaderBytes[8:12])[0]
    stdHeader['trace_number_in_original_field_record']=struct.unpack(INT,stdHeaderBytes[12:16])[0]
    stdHeader['energy_source_point_number']=struct.unpack(INT,stdHeaderBytes[16:20])[0]
    stdHeader['ensemble_number']=struct.unpack(INT,stdHeaderBytes[20:24])[0]
    stdHeader['trace_number_in_ensemble']=struct.unpack(INT,stdHeaderBytes[24:28])[0]
    stdHeader['trace_identification_code']=struct.unpack(SHORT,stdHeaderBytes[28:30])[0]
    stdHeader['number_of_vertically_stacked_traces']=struct.unpack(SHORT,stdHeaderBytes[30:32])[0]
    stdHeader['number_of_horizontally_stacked_traces']=struct.unpack(SHORT,stdHeaderBytes[32:34])[0]
    stdHeader['data_use']=struct.unpack(SHORT,stdHeaderBytes[34:36])[0]
    stdHeader['distance_center_source_to_center_of_receiver_group']=struct.unpack(INT,stdHeaderBytes[36:40])[0]
    stdHeader['elevation_of_receiver_group']=struct.unpack(INT,stdHeaderBytes[40:44])[0]
    stdHeader['surface_elevation_at_source']=struct.unpack(INT,stdHeaderBytes[44:48])[0]
    stdHeader['source_depth_below_surface']=struct.unpack(INT,stdHeaderBytes[48:52])[0]
    stdHeader['seismic_datum_elevation_at_group']=struct.unpack(INT,stdHeaderBytes[52:56])[0]
    stdHeader['seismic_datum_elevation_at_source']=struct.unpack(INT,stdHeaderBytes[56:60])[0]
    stdHeader['water_column_height_at_source']=struct.unpack(INT,stdHeaderBytes[60:64])[0]
    stdHeader['water_column_height_at_group']=struct.unpack(INT,stdHeaderBytes[64:68])[0]
    stdHeader['scalar_applied_to_elevations_and_depths']=struct.unpack(SHORT,stdHeaderBytes[68:70])[0]
    stdHeader['scalar_applied_to_coordinates']=struct.unpack(SHORT,stdHeaderBytes[70:72])[0]
    stdHeader['source_x']=struct.unpack(INT,stdHeaderBytes[72:76])[0]
    stdHeader['source_y']=struct.unpack(INT,stdHeaderBytes[76:80])[0]
    stdHeader['receiver_x']=struct.unpack(INT,stdHeaderBytes[80:84])[0]
    stdHeader['receiver_y']=struct.unpack(INT,stdHeaderBytes[84:88])[0]
    stdHeader['coordinate_units']=struct.unpack(SHORT,stdHeaderBytes[88:90])[0]
    stdHeader['weathering_velocity']=struct.unpack(SHORT,stdHeaderBytes[90:92])[0]
    stdHeader['subweathering_velocity']=struct.unpack(SHORT,stdHeaderBytes[92:94])[0]
    stdHeader['uphole_time_at_source']=struct.unpack(SHORT,stdHeaderBytes[94:96])[0]
    stdHeader['uphole_time_at_group']=struct.unpack(SHORT,stdHeaderBytes[96:98])[0]
    stdHeader['source_static_correction']=struct.unpack(SHORT,stdHeaderBytes[98:100])[0]
    stdHeader['group_static_correction']=struct.unpack(SHORT,stdHeaderBytes[100:102])[0]
    stdHeader['total_static_applied']=struct.unpack(SHORT,stdHeaderBytes[102:104])[0]
    stdHeader['lag_time_A']=struct.unpack(SHORT,stdHeaderBytes[104:106])[0]
    stdHeader['lag_time_B']=struct.unpack(SHORT,stdHeaderBytes[106:108])[0]
    stdHeader['delay_recording_time']=struct.unpack(SHORT,stdHeaderBytes[108:110])[0]
    stdHeader['mute_time_start']=struct.unpack(SHORT,stdHeaderBytes[110:112])[0]
    stdHeader['mute_time_end']=struct.unpack(SHORT,stdHeaderBytes[112:114])[0]
    stdHeader['number_of_samples']=struct.unpack(SHORT,stdHeaderBytes[114:116])[0]
    stdHeader['sample_interval']=struct.unpack(SHORT,stdHeaderBytes[116:118])[0]
    stdHeader['gain_type']=struct.unpack(SHORT,stdHeaderBytes[118:120])[0]
    stdHeader['instrument_gain']=struct.unpack(SHORT,stdHeaderBytes[120:122])[0]
    stdHeader['instrument_initial_gain']=struct.unpack(SHORT,stdHeaderBytes[122:124])[0]
    stdHeader['correlated']=struct.unpack(SHORT,stdHeaderBytes[124:126])[0]
    stdHeader['sweep_frequency_at_start']=struct.unpack(SHORT,stdHeaderBytes[126:128])[0]
    stdHeader['sweep_frequency_at_end']=struct.unpack(SHORT,stdHeaderBytes[128:130])[0]
    stdHeader['sweep_length']=struct.unpack(SHORT,stdHeaderBytes[130:132])[0]
    stdHeader['sweep_type']=struct.unpack(SHORT,stdHeaderBytes[132:134])[0]
    stdHeader['sweep_taper_at_start']=struct.unpack(SHORT,stdHeaderBytes[134:136])[0]
    stdHeader['sweep_taper_at_end']=struct.unpack(SHORT,stdHeaderBytes[136:138])[0]
    stdHeader['sweep_taper_type']=struct.unpack(SHORT,stdHeaderBytes[138:140])[0]
    stdHeader['alias_filter_frequency']=struct.unpack(SHORT,stdHeaderBytes[140:142])[0]
    stdHeader['alias_filter_slope']=struct.unpack(SHORT,stdHeaderBytes[142:144])[0]
    stdHeader['notch_filter_frequency']=struct.unpack(SHORT,stdHeaderBytes[144:146])[0]
    stdHeader['notch_filter_slope']=struct.unpack(SHORT,stdHeaderBytes[146:148])[0]
    stdHeader['low_cut_frequency']=struct.unpack(SHORT,stdHeaderBytes[148:150])[0]
    stdHeader['high_cut_frequency']=struct.unpack(SHORT,stdHeaderBytes[150:152])[0]
    stdHeader['low_cut_slope']=struct.unpack(SHORT,stdHeaderBytes[152:154])[0]
    stdHeader['high_cut_slope']=struct.unpack(SHORT,stdHeaderBytes[154:156])[0]
    stdHeader['year']=struct.unpack(SHORT,stdHeaderBytes[156:158])[0]
    stdHeader['day']=struct.unpack(SHORT,stdHeaderBytes[158:160])[0]
    stdHeader['hour']=struct.unpack(SHORT,stdHeaderBytes[160:162])[0]
    stdHeader['minute']=struct.unpack(SHORT,stdHeaderBytes[162:164])[0]
    stdHeader['second']=struct.unpack(SHORT,stdHeaderBytes[164:166])[0]
    stdHeader['time_basis_code']=struct.unpack(SHORT,stdHeaderBytes[166:168])[0]
    stdHeader['trace_weighting_factor']=struct.unpack(SHORT,stdHeaderBytes[168:170])[0]
    stdHeader['roll_switch_position_one']=struct.unpack(SHORT,stdHeaderBytes[170:172])[0]
    stdHeader['first_trace_in_original']=struct.unpack(SHORT,stdHeaderBytes[172:174])[0]
    stdHeader['last_trace_in_original']=struct.unpack(SHORT,stdHeaderBytes[174:176])[0]
    stdHeader['gap_size']=struct.unpack(SHORT,stdHeaderBytes[176:178])[0]
    stdHeader['over_travel']=struct.unpack(SHORT,stdHeaderBytes[178:180])[0]
    stdHeader['ensemble_x']=struct.unpack(INT,stdHeaderBytes[180:184])[0]
    stdHeader['ensemble_y']=struct.unpack(INT,stdHeaderBytes[184:188])[0]
    stdHeader['inline_number']=struct.unpack(INT,stdHeaderBytes[188:192])[0]
    stdHeader['crossline_number']=struct.unpack(INT,stdHeaderBytes[192:196])[0]
    stdHeader['shotpoint_number']=struct.unpack(INT,stdHeaderBytes[196:200])[0]
    stdHeader['scalar_for_shotpoint_number']=struct.unpack(SHORT,stdHeaderBytes[200:202])[0]
    stdHeader['trace_value_units']=struct.unpack(SHORT,stdHeaderBytes[202:204])[0]
    stdHeader['transduction_constant_mantissa']=struct.unpack(INT,stdHeaderBytes[204:208])[0]
    stdHeader['transduction_constant_exponent']=struct.unpack(SHORT,stdHeaderBytes[208:210])[0]
    stdHeader['transduction_units']=struct.unpack(SHORT,stdHeaderBytes[210:212])[0]
    stdHeader['device_trace_identifier']=struct.unpack(SHORT,stdHeaderBytes[212:214])[0]
    stdHeader['scalar_applied_to_times']=struct.unpack(SHORT,stdHeaderBytes[214:216])[0]
    stdHeader['source_type']=struct.unpack(SHORT,stdHeaderBytes[216:218])[0]
    stdHeader['source_energy_direction_vertical']=struct.unpack(SHORT,stdHeaderBytes[218:220])[0]
    stdHeader['source_energy_direction_crossline']=struct.unpack(SHORT,stdHeaderBytes[220:222])[0]
    stdHeader['source_energy_direction_inline']=struct.unpack(SHORT,stdHeaderBytes[222:224])[0]
    stdHeader['source_measurement_mantissa']=struct.unpack(INT,stdHeaderBytes[224:228])[0]
    stdHeader['source_measurement_exponent']=struct.unpack(SHORT,stdHeaderBytes[228:230])[0]
    stdHeader['source_measurement_unit']=struct.unpack(SHORT,stdHeaderBytes[230:232])[0]
    return stdHeader

'''
 data_format : support for most of the SEG-Y data formats
'''
def data_format(header, formats):
    itype = header['data_sample_format']
    ENDIAN = formats['ENDIAN']
    SAMPLE_TYPE = list(">f")
    BYTES = 4
    bFail = True
    if ENDIAN=='little':
        SAMPLE_TYPE[0]='<'
    if itype==1:
        # 4-byte IBM floating-point
        # ibm2ieee(sort.unpack('>L', f.read(4)))
        SAMPLE_TYPE[1]='L'
        BYTES = 4
        bFail=False
    if itype==2:
        # 4-byte two's complement integer
        SAMPLE_TYPE[1]='i'
        BYTES = 4
        bFail=False
    if itype==3:
        # 2-byte two's complement integer
        SAMPLE_TYPE[1]='h'
        BYTES = 2
        bFail=False
    if itype==4:
        print('4-byte fixed-point with gain (obsolete) - is unsupported')
        BYTES = 4
        bFail=True
    if itype==5:
        # 4-byte IEEE floating-point
        SAMPLE_TYPE[1]='f'
        BYTES = 4
        bFail=False
    if itype==6:
        # 8-byte IEEE floating-point
        SAMPLE_TYPE[1]='d'
        BYTES = 8
        bFail=False
    if itype==7:
        print('3-byte twos complement integer - is unsupported')
        BYTES = 3
        bFail=True
    if itype==8:
        # 1-byte, twos complement integer
        SAMPLE_TYPE[1]='b'
        BYTES = 1
        bFail=False
    if itype==9:
        # 8-byte, twos complement
        SAMPLE_TYPE[1]='q'
        BYTES = 8
        bFail=False
    if itype==10:
        # 4-byte, unsigned integer
        SAMPLE_TYPE[1]='I'
        BYTES = 4
        bFail=False
    if itype==11:
        # 2-byte, unsigned integer
        SAMPLE_TYPE[1]='H'
        BYTES = 2
        bFail=False
    if itype==12:
        # 8-byte, unsigned integer
        SAMPLE_TYPE[1]='Q'
        BYTES = 8
        bFail=False
    if itype==15:
        print("3-byte unsigned integer - not supported")
        BYTES = 3
        bFail=True
    if itype==16:
        # 1-byte unsigned integer
        SAMPLE_TYPE[1]='B'
        BYTES = 1
        bFail=False
    sample_type = "".join(SAMPLE_TYPE)
    #print(bFail,sample_type)

    return sample_type, BYTES


    
'''This script takes a SEGY flat file and turns it into processable chunks
The headers become JSON objects
It implements the Rev 2 standard from
https://seg.org/Portals/0/SEG/News%20and%20Resources/Technical%20Standards/seg_y_rev2_0-mar2017.pdf
'''
def SEGYingest(segyfilename, outdir, overwrite=True):
    directory_services.makedir(outdir)
    headerDir = os.path.join(outdir,'json')
    directory_services.makedir(headerDir)
    tmpdir = os.path.join(outdir,'tmp')
    directory_services.makedir(tmpdir)
    
    
    # The 1-second blocks of data are memory mapped arrays...
    memmapList = []
    tmpList = []

    # Open the file for binary read
    rawObj=open(segyfilename,'rb')

    # hard-coded constants - from the standard
    FULL_HEADER_BYTES = 3600
    STANDARD_TRACE_HEADER_SIZE = 240

    # interpret the file header
    fullHeader = rawObj.read(FULL_HEADER_BYTES)
    header, formats = file_header(fullHeader)

    
    ntraces = header['number_of_traces_in_file']
    itrace=0
    # Local variables for convenience
    nx = header['number_of_traces_in_file']
    nt = header['extended_samples_per_trace']
    prf = int(1000000.0/header['sample_interval'])
    print('nx',nx)
    print('nt',nt)
    print('prf',prf)
    #print (json.dumps(header, sort_keys=True, indent=4, separators=(',', ': ')))
    xaxis = numpy.zeros((nx,1),dtype=numpy.double)
    ntOld=nt

    
    # read each trace
    for a in range(ntraces):
        nt=ntOld
        #print(a)
        stdHeaderBytes = rawObj.read(STANDARD_TRACE_HEADER_SIZE)
        stdHeader= trace_header(stdHeaderBytes, formats)
        # 233-240 : either binary zeros or the eight character trace header name SEG00000.
        #           May be ASCII or EBCDIC text.

        #if header['max_additional_240byte_headers']>0:
        #    print("additional trace headers found")
        # NOTE: this is wrong... should be stdHeader['scalar_applied_to_elevations_and_depths']...

        # RISK OF DIVISION BY ZERO HERE...SO AVOID BY SETTING 0 TO 1
        # IT's a short integer...
        if stdHeader['scalar_applied_to_coordinates']<1:
            stdHeader['scalar_applied_to_coordinates']=1
        xaxis[a]=(stdHeader['elevation_of_receiver_group']/stdHeader['scalar_applied_to_coordinates'])

        # determine the format of the numbers
        sample_type, BYTES = data_format(header, formats)

        nsamps = header['number_of_samples_per_data_trace']
        nsampsBig = header['extended_samples_per_trace']
        if (nsampsBig>nsamps):
            nsamps=nsampsBig
        traceBytes = rawObj.read(nsamps*BYTES)
        # interpret trace
        trace = numpy.frombuffer(traceBytes,dtype=sample_type)
        if trace.shape[0]<nt:
            nt=trace.shape[0]
        # Have to support itype==1 - this is a bit clunky so look at this for refactoring
        itype = header['data_sample_format']
        if itype==1:
            trace = numpy.double(ibm2ieee(trace))    
        if a==0:
            #---RESTART--- when doing lots of SEGYs in parallel the compute system might fail
            #              so we can do a quick check to see whether all the 1-second files for t
            #              this SEGY already exist            
            datestring = str(stdHeader['year'])+' '+str(stdHeader['day']).zfill(3)+' '+str(stdHeader['hour']).zfill(2)+':'+str(stdHeader['minute']).zfill(2)+':'+str(stdHeader['second']).zfill(2)
            print(datestring)
            print('time_basis_code:',str(stdHeader['time_basis_code']))
            datestamp = datetime.datetime.strptime(datestring,'%Y %j %H:%M:%S')
            unixtime = 0
            try:
                unixtime = int(datestamp.timestamp())
            except:
                pass
            bDone = True
            nsecs = 0
            for mapid in range(0,nt,prf):
                if not os.path.exists(os.path.join(outdir,str(unixtime+nsecs)+'.npy')):
                    bDone = False
                nsecs += 1
            if (bDone==True):
                print('skipping ', segyfilename, ' as already ingested')
                rawObj.close()
                return
            #---END OF RESTART----------------------------------------------
            # first time through... so we write the full header
            headerDir = os.path.join(headerDir,str(unixtime))
            if not os.path.exists(headerDir):
                os.makedirs(headerDir)
            with open(os.path.join(headerDir,str(unixtime)+'.json'),'w') as outJson:
                outJson.write(json.dumps(header, sort_keys=True, indent=4, separators=(',', ': ')))
            # create a memap list (because the SEGY is the 'wrong' way around...
            nsecs=0
            for mapid in range(0, nt, prf):
                fileout = os.path.join(tmpdir,str(unixtime+nsecs)+'.npy')
                #fileout = os.path.join(outdir,str(unixtime+nsecs)+'.npy')
                print(fileout)
                memmapList.append(numpy.memmap(fileout, dtype=numpy.double, mode='w+', shape=(nx,prf)))
                tmpList.append(fileout)
                nsecs=nsecs+1
        # Write this data to all the memmapped files
        nsecs = 0
        #print("AZURE-issue: ",nt,prf,trace.shape)
        for mapid in range(0,nt,prf):
            (memmapList[nsecs])[a,0:prf] = trace[mapid:mapid+prf]
            nsecs = nsecs+1
        # Write out the header at the end - because the directory is not set until we know about unixtimestamp
        # We need pretty print, even though we are writing to a stream, because we expect human inspection of the headers.
        with open(os.path.join(headerDir,str(a)+'.json'),'w') as outJson:
            outJson.write(json.dumps(stdHeader, sort_keys=True, indent=4, separators=(',', ': ')))
    for mapped in memmapList:        
        mapped.flush()
    # convert to numpy.save() so that we get metadata...
    nsecs=0
    print('Converting temporary files to permanent')
    for mapped in memmapList:
        fileout = os.path.join(outdir,str(unixtime+nsecs)+'.npy')
        print(fileout)
        numpy.save(fileout, numpy.array(mapped))
        nsecs=nsecs+1
        del mapped
        # override lazy behaviour...
        mapped = None
    memmapList = None
    # remove temporary files...
    for filename in tmpList:
        os.remove(filename)
    # Assumption that the fibre depth points do not change over the measurement period
    xaxisfilename = os.path.join(outdir,'measured_depth.npy')
    if not os.path.exists(xaxisfilename):
        numpy.save(xaxisfilename, xaxis)
    rawObj.close()



# TEST CODE...
def main():
    # A test of the SEGYingest...
    segyfilename = "{PATH}\\SegyPhase\\demo_003_F0086_S20180828_140213.862+0000_0000.sgy"
    outdir = "{PATH}\\sgy"

    SEGYingest(segyfilename, outdir, overwrite=False)
    # For google cloud use tensorflow for numpy.save and numpy.load... possibly restricted to float32...
    # https://stackoverflow.com/questions/41633748/load-numpy-array-in-google-cloud-ml-job

if __name__ == "__main__":
    main()



# (C) 2020, Schlumberger. Refer to LICENSE

import numpy
import matplotlib.pyplot as plt
import datetime
import scipy.signal
import os
import distpy.io_help.io_helpers as io_helpers
import distpy.io_help.directory_services as directory_services
import distpy.calc.extra_numpy as extra_numpy
import distpy.calc.extra_pyplot as extra_pyplot
import distpy.calc.unit_handler as unit_handler
import distpy.calc.pub_command_set as pub_command_set
import distpy.calc.processing_commands as processing_commands

'''
 read_zones : read a CSV file containing measured_depth zone information
              for plot annotation
'''
def read_zones(csvFile):
    lines = directory_services.csv_reader(csvFile)
    #conversion = 1.0
    #if (lines[1][0]=="ft"):
    #    conversion = FT_TO_M
    zones=[]
    for a in range(2, len(lines),2):
        one_zone={}
        tokens = lines[a]
        one_zone['start'] = float(tokens[0])
        one_zone['name'] = tokens[-1]
        tokens = lines[a+1]
        one_zone['end'] = float(tokens[0])
        zones.append(one_zone)
    print(len(zones))
    return zones

'''
 plotgenerator : generates a command_list which is then executed to generate the
                 plots. 
'''
def plotgenerator(dirin, dirout, plotData):
    
    # Configure the hardware
    boxsize = plotData.get('BOXSIZE', 500)
    extra_numpy.set_boxsize(boxsize)

    depth_display_unit = plotData['depth_display_unit']
    start_of_fibre = plotData['start_of_fibre']
    end_of_fibre = plotData['end_of_fibre']
    
    figure_size = plotData['figure_size']
    dots_per_inch = plotData['dpi']
    event_list = plotData['label_list']
    
    ## clusters and stages
    segs_blob = plotData['well_segments']
    
    # blob locations
    TIME_REF_BLOB = directory_services.path_join(dirin,plotData['time_reference'])
    DEPTH_REF_BLOB = directory_services.path_join(dirin,plotData['depth_reference'])
    

    time_ref = directory_services.load(TIME_REF_BLOB)
    time_list = extra_pyplot.time_stamps(time_ref)
    nt = time_ref.shape[0]
    
    depth_ref = directory_services.load(DEPTH_REF_BLOB)
    if (depth_display_unit=="ft"):
        depth_ref = unit_handler.M_TO_FT(depth_ref)
    nx = depth_ref.shape[0]

    # well segmentation for flow allocations
    well_segs = read_zones(segs_blob)

    # same command factory as for strain-rate processing - giving maths functions...
    #dir_suffix = plotData.get('directory_out',dirout)
    #if not dir_suffix=='NONE':
    #   dirval = os.path.join(dirout,dir_suffix)
    dirval = dirout
    directory_services.makedir(dirval)
    
    plt.switch_backend('Agg')
    command_list = []
    # small 2D array for a command zero
    data=numpy.zeros((10,10),dtype=numpy.double)
    command_list.append(pub_command_set.DataLoadCommand(data,{})) 

    for plot in plotData['plots']:
        # Locally package the global information for the generation of this particular plot
        plot['nx']=nx
        plot['nt']=nt
        plot['label_list']= plotData['label_list']
        plot['time_ref']=time_ref
        plot['depth_ref']=depth_ref
        plot['directory_in']=dirin
        
        plot['depth_display_unit'] = plotData['depth_display_unit']
        plot['start_of_fibre'] = plotData['start_of_fibre']
        plot['end_of_fibre'] = plotData['end_of_fibre']
        plot['figure_size'] = plotData['figure_size']
        plot['dpi'] = plotData['dpi']
        plot['well_segments'] = well_segs
        # internal mapping to the commands
        # 1. commands have names
        # 2. in_uid from a previous command (here always command zero)
        plot['name'] = plot['plot_type']
        plot['in_uid']=0

        plot['directory_out']=dirval
        command_list.append(processing_commands.CommandFactory(command_list,plot))


    # Actual plot generation occurs here...
    for command in command_list:
        print(command)
        command.execute()



# (C) 2020, Schlumberger. Refer to LICENSE.
'''
 cupy_facade - a do-nothing class for devices that don't have cupy installed

 This allows CPU/GPU agnostic code to be written. We engage with the GPU
 via the cuda_facade.
'''
import numpy

def get_array_module(x):
    return numpy


def asarray(x):
    return x

def asnumpy(x):
    return x



# (C) 2020, Schlumberger. Refer to LICENSE

########################################
# USER CONFIGURABLE VALUES:
jsonSystemConfig = "{PATH}\\systemh5Config.json"
########################################
import argparse

import os
import multiprocessing
import distpy.io_help.io_helpers as io_helpers
import distpy.io_help.h5_helpers as h5_helpers
import distpy.io_help.directory_services as ds

def main(configOuterFile):
    # standard configuration information
    basedir,dirout,configFile,PARALLEL,NCPU,BOX_SIZE, xaxisfile,taxisfile,prf = io_helpers.systemConfig(configOuterFile)
    ds.makedir(dirout)

    print(basedir)
    print(dirout)
    #scan for directories
    h5_files=[]
    for root, dirs, files in os.walk(basedir):
        for datafile in files:
            h5_files.append(os.path.join(root,datafile))

    if not PARALLEL:
        for fname in h5_files:
            print(fname)
            h5_helpers.ingest_h5(fname, dirout)
    else:
        manager = multiprocessing.Manager()
        q = manager.Queue()
        print(multiprocessing.cpu_count())
        
        pool = multiprocessing.Pool(processes=NCPU, maxtasksperchild=1)

        jobs = []
        for fname in h5_files:
            job = pool.apply_async(h5_helpers.ingest_h5, [fname,dirout])
            print(job)
            jobs.append(job)

        for job in jobs:
            job.get()
            print(job)

        q.put('kill')
        pool.close()

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-f","--filename", help='A configuration file')
    args = parser.parse_args()
    main(args.filename)



# (C) 2020, Schlumberger. Refer to LICENSE

#
# witsmlfbe : handles reading and writing of the WITSML FBE format
#             and also contains tools for converting between LAS 2.0
#             and WITSML
#

import xml.etree.ElementTree as ET
import os
import numpy
import datetime
from os import listdir
from os.path import isfile, join
from xml.dom import minidom

##############################################################################
# CONFIGURATION
#################################
inpath='{PATH}/LAS/'
outpath='{PATH}/FBE/'
#################################


##############################################################################
# CONSTANTS
FAILED_MESSGE = 'Expecting a LAS 2.0 file with FBE data'
#LAS 2.0 hard coded numbers...
#0123456789012345678901234567890123
#DATE. 2016-12-13 07:13:49+00:00 : DATE
LAS2_DATESTART = 6
LAS2_DATEEND = 32
LAS2_ENDOFDESC = 33
LAS2_UNITSTART = 32
LAS2_UNITEND = 34
LAS2_DATE_TAG = 'DATE.'
LAS2_CURVE_TAG = '~Curve Information'
LAS2_PARAMS_TAG = '~Params'
LAS2_ASCII_TAG = '~ASCII'

# Sufficient specification for FBE in WITSML...
# common tags
SCHEMA = '{http://www.witsml.org/schemas/131}'
WITSML_NAME_TAG = SCHEMA + 'name'
NO_ATTR = {}
# LAS 2.0 does not tell us about the box or fiber used
UNKNOWN = 'UNKNOWN'
UNKNOWN_UID = {'uid': UNKNOWN}
UNKNOWN_UIDREF = {'uidRef': UNKNOWN}
TRUE = {'uid': '1'}
# There is no unique identifier for the LAS 2.0 data so we use this fixed dummy value
WELLLOG_UNIQUE = '5B8B6FB8-AACA-11E8-9637-0CC47A31FD84'
WELLLOG_UIDREF = {'uidRef': WELLLOG_UNIQUE}
# dummy well name
WELLNAME = 'FBR01'

ROOT_TAG = '{http://www.witsml.org/schemas/131}WITSMLComposite'
ROOT_ATTR = {'version': '1.3.1.1', '{http://www.w3.org/2001/XMLSchema-instance}schemaLocation': 'http://www.witsml.org/schemas/131 WITSML_composite.xsd'}
#...  tag    =  {http://www.witsml.org/schemas/131}wellSet
#...  attrib =  {}
#...  text   =  
WELLSET_TAG = SCHEMA + 'wellSet'
WELLSET_ATTR = NO_ATTR
#...|...  tag    =  {http://www.witsml.org/schemas/131}well
#...|...  attrib =  {'uid': 'FBR01'}
#...|...  text   =  
WELL_TAG = SCHEMA + 'well'
WELL_ATTR = {'uid': WELLNAME}
#...|...|...  tag    =  {http://www.witsml.org/schemas/131}name
#...|...|...  attrib =  {}
#...|...|...  text   =  FBR01
#...|...|...  tag    =  {http://www.witsml.org/schemas/131}wellboreSet
#...|...|...  attrib =  {}
#...|...|...  text   =  
WELL_NAME_TAG = WITSML_NAME_TAG
WELL_NAME_ATTR = NO_ATTR
WELL_NAME_TEXT = WELLNAME
WELLBORESET_TAG = SCHEMA + 'wellboreSet'
WELLBORESET_ATTR = NO_ATTR
#...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}wellbore
#...|...|...|...  attrib =  {'uid': 'FBR01'}
#...|...|...|...  text   =  
WELLBORE_TAG = SCHEMA + 'wellbore'
WELLBORE_ATTR = WELL_ATTR
#...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}name
#...|...|...|...|...  attrib =  {}
#...|...|...|...|...  text   =  FBR01
#...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}dtsInstalledSystemSet
#...|...|...|...|...  attrib =  {}
#...|...|...|...|...  text   =  
WELLBORE_NAME_TAG = WELL_NAME_TAG
WELLBORE_NAME_ATTR = WELL_NAME_ATTR
WELLBORE_NAME_TEXT = WELLNAME
# this is for DTS...re-used for DAS
DTSINSTALLEDSYSTEMSET_TAG = SCHEMA + 'dtsInstalledSystemSet'
DTSINSTALLEDSYSTEMSET_ATTR = NO_ATTR
#...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}dtsInstalledSystem
#...|...|...|...|...|...  attrib =  {'uid': '0C-C4-7A-31-FD-84'}
#...|...|...|...|...|...  text   =



DTSINSTALLEDSYSTEM_TAG = SCHEMA + 'dtsInstalledSystem'
#...we don't know the exact box in this case because it is not recorded in LAS 2.0
DTSINSTALLEDSYSTEM_ATTR = UNKNOWN_UID

# for the DTS Installed System Set branch
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}name
#...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...  text   =  hDvs_01
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}fiberInformation
#...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...  text   =  
DTSINSTALLEDSYSTEM_NAME_TAG = WITSML_NAME_TAG
DTSINSTALLEDSYSTEM_NAME_ATTR = NO_ATTR
DTSINSTALLEDSYSTEM_NAME_TEXT = UNKNOWN
FIBERINFORMATION_TAG = SCHEMA + 'fiberInformation'
FIBERINFORMATION_ATTR = NO_ATTR

#...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}fiber
#...|...|...|...|...|...|...|...  attrib =  {'uid': '1'}
#...|...|...|...|...|...|...|...  text   =  
FIBER_TAG = SCHEMA + 'fiber'
FIBER_ATTR = TRUE
#...|...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}name
#...|...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...|...  text   =  FBR01
#...|...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}mode
#...|...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...|...  text   =  singlemode
FIBER_NAME_TAG = WITSML_NAME_TAG
FIBER_NAME_ATTR = NO_ATTR
FIBER_NAME_TEXT = WELLNAME
FIBERMODE_TAG = SCHEMA + 'mode'
FIBERMODE_ATTR = NO_ATTR
FIBERMODE_TEXT = UNKNOWN

#...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}dtsMeasurementSet
#...|...|...|...|...  attrib =  {}
#...|...|...|...|...  text   =  
DTSMEASUREMENTSET_TAG = SCHEMA + 'dtsMeasurementSet'
DTSMEASUREMENTSET_ATTR = NO_ATTR
#...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}dtsMeasurement
#...|...|...|...|...|...  attrib =  {'uid': '5B8B6FB8-AACA-11E8-9637-0CC47A31FD84'}
#...|...|...|...|...|...  text   =  
DTSMEASUREMENT_TAG = SCHEMA + 'dtsMeasurement'
DTSMEASUREMENT_ATTR = UNKNOWN_UID
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}name
#...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...  text   =  INT_FBE_SIG
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}installedSystemUsed
#...|...|...|...|...|...|...  attrib =  {'uidRef': '0C-C4-7A-31-FD-84'}
#...|...|...|...|...|...|...  text   =  hDvs_01
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}dataInWellLog
#...|...|...|...|...|...|...  attrib =  {'uidRef': '5B8B6FB8-AACA-11E8-9637-0CC47A31FD84'}
#...|...|...|...|...|...|...  text   =  5B8B6FB8-AACA-11E8-9637-0CC47A31FD84
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}connectedToFiber
#...|...|...|...|...|...|...  attrib =  {'uidRef': '1'}
#...|...|...|...|...|...|...  text   =  FBR01
DTSMEASUREMENT_NAME_TAG = WITSML_NAME_TAG
DTSMEASUREMENT_NAME_ATTR = NO_ATTR
DTSMEASUREMENT_NAME_TEXT = UNKNOWN
INSTALLEDSYSTEMUSED_TAG = SCHEMA + 'installedSystemUsed'
INSTALLEDSYSTEMUSED_ATTR = UNKNOWN_UIDREF
INSTALLEDSYSTEMUSED_TEXT = UNKNOWN
DATAINWELLLOG_TAG = SCHEMA + 'dataInWellLog'
DATAINWELLLOG_ATTR = WELLLOG_UIDREF
DATAINWELLOG_TEXT = WELLLOG_UNIQUE
#...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}wellLogSet
#...|...|...|...|...  attrib =  {}
#...|...|...|...|...  text   =  
WELLLOGSET_TAG = SCHEMA + 'wellLogSet'
WELLLOGSET_ATTR = NO_ATTR
#...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}wellLog
#...|...|...|...|...|...  attrib =  {'uid': '5B8B6FB8-AACA-11E8-9637-0CC47A31FD84'}
#...|...|...|...|...|...  text   = 
# wellLogSet
WELLLOG_TAG = SCHEMA + 'wellLog'
WELLLOG_ATTR = WELLLOG_UIDREF
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}name
#...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...  text   =  5B8B6FB8-AACA-11E8-9637-0CC47A31FD84
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}serviceCompany
#...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...  text   =  Schlumberger
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}creationDate
#...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...  text   =  2018-08-28T13:57:13+00:00
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}indexType
#...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...  text   =  depth
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}nullValue
#...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...  text   =  NULL
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}logCurveInfo
#...|...|...|...|...|...|...  attrib =  {'uid': '1'}
#...|...|...|...|...|...|...  text   =
# wellLog
WELLLOG_NAME_TAG = WITSML_NAME_TAG
WELLLOG_NAME_ATTR = NO_ATTR
WELLLOG_NAME_TEXT = WELLLOG_UNIQUE
SERVICECOMPANY_TAG = SCHEMA + 'serviceCompany'
SERVICECOMPANY_ATTR = NO_ATTR
SERVICECOMPANY_NAME = UNKNOWN
CREATIONDATE_TAG = SCHEMA + 'creationDate'
CREATIONDATE_ATTR = NO_ATTR
##############...LAS 2.0 data in here###########
CREATIONDATE_TEXT = '2018-08-28T13:57:13+00:00'
INDEXTYPE_TAG = SCHEMA + 'indexType'
INDEXTYPE_ATTR = NO_ATTR
INDEXTYPE_TEXT = 'depth'
NULLVALUE_TAG = SCHEMA + 'nullValue'
NULLVALUE_ATTR = NO_ATTR
NULLVALUE_TEXT = 'NULL'
# AT THIS POINT WE GET INTO LIST OF INFORMATION THAT LAS 2.0 provides...
LOGCURVEINFO_TAG = SCHEMA + 'logCurveInfo'
LOGCURVEINFO_ATTR = TRUE
#...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}mnemonic
#...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...  text   =  DEPTH
#...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}unit
#...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...  text   =  m
#...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}curveDescription
#...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...  text   =  depth along the fiber
MNEMONIC_TAG = SCHEMA + 'mnemonic'
MNEMONIC_ATTR = NO_ATTR
UNIT_TAG = SCHEMA + 'unit'
UNIT_ATTR = NO_ATTR
CURVEDESCRIPTION_TAG = SCHEMA + 'curveDescription'
CURVEDESCRIPTION_ATTR = NO_ATTR
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}blockInfo
#...|...|...|...|...|...|...  attrib =  {'uid': '1'}
#...|...|...|...|...|...|...  text   =  
BLOCKINFO_TAG = SCHEMA + 'blockInfo'
BLOCKINFO_ATTR = TRUE
#...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}indexType
#...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...  text   =  length
#...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}direction
#...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...  text   =  increasing
#...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}indexCurve
#...|...|...|...|...|...|...|...  attrib =  {'columnIndex': '1'}
#...|...|...|...|...|...|...|...  text   =  LENGTH
#...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}blockCurveInfo
#...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...  text   =
INDEXTYPE_TAG = SCHEMA + 'indexType'
INDEXTYPE_ATTR = NO_ATTR
DIRECTION_TAG = SCHEMA + 'direction'
DIRECTION_ATTR = NO_ATTR
INDEXCURVE_TAG = SCHEMA + 'indexCurve'
INDEXCURVE_ATTR = {'columnIndex': '1'}
BLOCKCURVEINFO_TAG = SCHEMA + 'blockCurveInfo'
BLOCKCURVEINFO_ATTR = NO_ATTR
#...|...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}curveId
#...|...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...|...  text   =  1
#...|...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}columnIndex
#...|...|...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...|...|...  text   =  1
CURVEID_TAG = SCHEMA + 'curveId'
CURVEID_ATTR = NO_ATTR
COLUMNINDEX_TAG = SCHEMA + 'columnIndex'
COLUMNINDEX_ATTR = NO_ATTR
#...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}logData
#...|...|...|...|...|...|...  attrib =  {}
#...|...|...|...|...|...|...  text   =  
LOGDATA_TAG = SCHEMA + 'logData'
LOGDATA_ATTR = NO_ATTR
#...|...|...|...|...|...|...|...  tag    =  {http://www.witsml.org/schemas/131}data
#...|...|...|...|...|...|...|...  attrib =  {'id': '1'}
#...|...|...|...|...|...|...|...  text   =  -2.02918,574.19,98.1762,25.1918,17.0701,27.9438
DATA_TAG = SCHEMA + 'data'
DATA_ATTR = {'id': '1'}

CONNECTEDTOFIBER_TAG = SCHEMA + 'connectedToFiber'
CONNECTEDTOFIBER_ATTR = TRUE

##################################################################################################
# FUNCTIONS:
##for testing
'''
 printsubbranch : used in testing whether the formed XML is correct
'''
def printsubbranch(elements, dots):
    for elem in elements:
        print(dots,' tag    = ',elem.tag)
        print(dots,' attrib = ',elem.attrib)
        print(dots,' text   = ',elem.text)
        more_elements = elem.findall("./")
        if len(more_elements)>0:
            printsubbranch(more_elements, dots+'|...')

'''
  wrong_file_type : only LAS 2.0 is supported
'''
def wrong_file_type(lines):
    if lines[1][:11] != 'VERS.   2.0':
        return True    
    return False

'''
 lineIndex : determine the line number in the original data, corresponding to
             a given XML tab
'''
def lineIndex(lines,tag):
    test_len=len(tag)
    icount=0
    for line in lines:
        if line[:test_len] == tag:
            return icount,line
        icount+=1
    return 1,'none'

'''
 generic_curves : WITSML headers for the FBE style (band-limited data)
'''
def generic_curves(low_freq,high_freq, prf):
    curves = []
    #band-00
    curve={}
    curve['mnemonic'] = 'band_0-nyquist'
    curve['unit'] = 'unitless'
    curve['description'] = 'from 0 Hz to Nyquist Hz (prf='+str(prf)+ ')'
    curves.append(curve)

    for a in range(len(low_freq)):
        curve={}
        if a<9:
            curve['mnemonic'] = 'band_0' +str(a+1)+'_'+str(low_freq[a])+'-'+str(high_freq[a])
        else:
            curve['mnemonic'] = 'band_' +str(a+1)+'_'+str(low_freq[a])+'-'+str(high_freq[a])
        curve['unit'] = 'unitless'
        curve['description'] = 'from '+str(low_freq[a])+ ' Hz to ' +str(high_freq[a])+' Hz (prf='+str(prf) + ')'
        curves.append(curve)
    return curves


'''
 set_unkown_constants : allows over-writing of the defaults 
'''
def set_unknown_constants(data_style):
    global UNKNOWN, UNKNOWN_UID, UNKNOWN_UIDREF, DTSINSTALLEDSYSTEM_ATTR, DTSINSTALLEDSYSTEM_NAME_TEXT
    global FIBERMODE_TEXT, DTSMEASUREMENT_ATTR, DTSMEASUREMENT_NAME_TEXT, INSTALLEDSYSTEMUSED_ATTR
    global INSTALLEDSYSTEMUSED_TEXT, SERVICECOMPANY_NAME
    # The option to update the UNKNOWN to be something more useful
    UNKNOWN = data_style
    UNKNOWN_UID = {'uid': data_style}
    UNKNOWN_UIDREF = {'uidRef': data_style}
    DTSINSTALLEDSYSTEM_ATTR = UNKNOWN_UID
    DTSINSTALLEDSYSTEM_NAME_TEXT = UNKNOWN
    FIBERMODE_TEXT = UNKNOWN
    DTSMEASUREMENT_ATTR = UNKNOWN_UID
    DTSMEASUREMENT_NAME_TEXT = UNKNOWN
    INSTALLEDSYSTEMUSED_ATTR = UNKNOWN_UIDREF
    INSTALLEDSYSTEMUSED_TEXT = UNKNOWN
    SERVICECOMPANY_NAME = UNKNOWN


'''
  las2fbe : Given a LAS file, create a WITSML format fbe for Techlog.

            NOTE: this is not a generic reader of the LAS format and it is also
                  not a generic writer for WITSML 1.3.1.1

            Only submit files in LAS which have the column format of
            depth followed by FBE bands. No other files are supported.
'''
def las2fbe(filename):
    # import the LAS as lines of text
    f=open(filename,"r")
    lines = f.readlines()
    f.close()

    root = ET.Element(ROOT_TAG,ROOT_ATTR)
    wellset = ET.SubElement(root, WELLSET_TAG, WELLSET_ATTR)
    well = ET.SubElement(wellset, WELL_TAG, WELL_ATTR)
    

    wellname = ET.SubElement(well, WELL_NAME_TAG, WELL_NAME_ATTR)
    wellname.text = WELL_NAME_TEXT
    
    wellboreSet = ET.SubElement(well, WELLBORESET_TAG, WELLBORESET_ATTR)
    # The following level down is the wellbore
    wellbore = ET.SubElement(wellboreSet, WELLBORE_TAG, WELLBORE_ATTR)
    # wellbore has 4 elements
    wellborename = ET.SubElement(wellbore, WELLBORE_NAME_TAG, WELLBORE_NAME_ATTR)
    wellborename.text = WELLBORE_NAME_TEXT
    dtsInstalledSystemSet = ET.SubElement(wellbore, DTSINSTALLEDSYSTEMSET_TAG, DTSINSTALLEDSYSTEMSET_ATTR)
    dtsMeasurementSet = ET.SubElement(wellbore, DTSMEASUREMENTSET_TAG, DTSMEASUREMENTSET_ATTR)
    wellLogSet = ET.SubElement(wellbore, WELLLOGSET_TAG, WELLLOGSET_ATTR)
    # ....dtsInstalledSystemSet has information on the DTS/DAS system used
    dtsInstalledSystem = ET.SubElement(dtsInstalledSystemSet, DTSINSTALLEDSYSTEM_TAG, DTSINSTALLEDSYSTEM_ATTR)
    # ....|....dtsInstalledSystem has information on the DTS/DAS system used
    dtsInstalledSystemName = ET.SubElement(dtsInstalledSystem, DTSINSTALLEDSYSTEM_NAME_TAG, DTSINSTALLEDSYSTEM_NAME_ATTR)
    dtsInstalledSystemName.text = DTSINSTALLEDSYSTEM_NAME_TEXT
    fiberInformation = ET.SubElement(dtsInstalledSystem, FIBERINFORMATION_TAG, FIBERINFORMATION_ATTR)
    # ....|....|....fiberInformation - including mode
    fiber = ET.SubElement(fiberInformation, FIBER_TAG, FIBER_ATTR)
    # ....|....|....|....fiber - the name and mode of the fiber
    fiberName = ET.SubElement(fiber, FIBER_NAME_TAG, FIBER_NAME_ATTR)
    fiberName.text = FIBER_NAME_TEXT
    mode = ET.SubElement(fiber, FIBERMODE_TAG, FIBERMODE_ATTR)
    mode.text = FIBERMODE_TEXT
    # ....dtsMeasurementSet
    dtsMeasurement = ET.SubElement(dtsMeasurementSet, DTSMEASUREMENT_TAG, DTSMEASUREMENT_ATTR)
    # ....|....dtsMeasurement
    dtsMeasurementName = ET.SubElement(dtsMeasurement, DTSMEASUREMENT_NAME_TAG,DTSMEASUREMENT_NAME_ATTR)
    dtsMeasurementName.text = DTSMEASUREMENT_NAME_TEXT
    installedSystemUsed = ET.SubElement(dtsMeasurement, INSTALLEDSYSTEMUSED_TAG,INSTALLEDSYSTEMUSED_ATTR)
    installedSystemUsed.text = INSTALLEDSYSTEMUSED_TEXT
    dataInWellLog = ET.SubElement(dtsMeasurement, DATAINWELLLOG_TAG, DATAINWELLLOG_ATTR)
    dataInWellLog.text = WELLLOG_UNIQUE
    connectedToFiber = ET.SubElement(dtsMeasurement, CONNECTEDTOFIBER_TAG, CONNECTEDTOFIBER_ATTR)
    # ....wellLogSet
    wellLog = ET.SubElement(wellLogSet, WELLLOG_TAG, WELLLOG_ATTR)
    # ....|....wellLog
    wellLogName = ET.SubElement(wellLog, WELLLOG_NAME_TAG, WELLLOG_NAME_ATTR)
    wellLogName.text = WELLLOG_NAME_TEXT
    serviceCompany = ET.SubElement(wellLog, SERVICECOMPANY_TAG, SERVICECOMPANY_ATTR)
    serviceCompany.text = SERVICECOMPANY_NAME
    creationDate = ET.SubElement(wellLog, CREATIONDATE_TAG, CREATIONDATE_ATTR)
    #############LAS 2.0 INFO IN HERE.......###################
    istart, line = lineIndex(lines,LAS2_DATE_TAG)
    creationDate.text = line[LAS2_DATESTART:LAS2_DATEEND]
    indexType = ET.SubElement(wellLog, INDEXTYPE_TAG, INDEXTYPE_ATTR)
    indexType.text = INDEXTYPE_TEXT
    nullValue = ET.SubElement(wellLog, NULLVALUE_TAG, NULLVALUE_ATTR)
    nullValue.text = NULLVALUE_TEXT

    istart, line = lineIndex(lines,LAS2_CURVE_TAG)
    iend, line = lineIndex(lines,LAS2_PARAMS_TAG)
    lognumber=1
    band=1
    # populate the column descriptions
    for line in lines[istart+1:iend-1]:
        logCurveInfo = ET.SubElement(wellLog, LOGCURVEINFO_TAG, {'uid' : str(lognumber)} )
        blockInfo = ET.SubElement(wellLog, BLOCKINFO_TAG, {'uid' : str(lognumber)})
        
        mnemonic = ET.SubElement(logCurveInfo, MNEMONIC_TAG, MNEMONIC_ATTR)
        unit = ET.SubElement(logCurveInfo, UNIT_TAG, UNIT_ATTR)
        curveDescription = ET.SubElement(logCurveInfo,CURVEDESCRIPTION_TAG, CURVEDESCRIPTION_ATTR)
        if line[:4] == 'DEPT':
            mnemonic.text = 'DEPTH'
            unit.text = line[LAS2_UNITSTART:LAS2_UNITEND]
            curveDescription.text = 'depth along the fiber'
            indexType = ET.SubElement(blockInfo, INDEXTYPE_TAG, INDEXTYPE_ATTR)
            indexType.text = 'length'
            direction = ET.SubElement(blockInfo, DIRECTION_TAG, DIRECTION_ATTR)
            direction.text = 'increasing'
            indexCurve = ET.SubElement(blockInfo, INDEXCURVE_TAG, INDEXCURVE_ATTR)
            indexCurve.text = str(int(1))
        else:
            textpart = 'BAND-'
            if band<10:
                textpart +='0'
            textpart+=str(int(band))
            mnemonic.text = textpart
            unit.text = 'unitless'
            curveDescription = 'Frequency Band Energy '+line[:LAS2_ENDOFDESC]
            blockCurveInfo = ET.SubElement(blockInfo, BLOCKCURVEINFO_TAG, BLOCKCURVEINFO_ATTR)
            curveId = ET.SubElement(blockCurveInfo, CURVEID_TAG, CURVEID_ATTR)
            curveId.text = str(int(band))
            columnIndex = ET.SubElement(blockCurveInfo, COLUMNINDEX_TAG, COLUMNINDEX_ATTR)
            columnIndex.text = curveId.text          
            band+=1
        lognumber+=1
    # populate the log data
    logData = ET.SubElement(wellLog, LOGDATA_TAG, LOGDATA_ATTR)
    istart, line = lineIndex(lines,LAS2_ASCII_TAG)
    for a in range(istart+1,len(lines)):
        line = lines[a]
        if len(line)>1:
            data = ET.SubElement(logData,DATA_TAG, DATA_ATTR)
            line = line.rstrip()
            line = line.lstrip()
            line = line.replace('  ',' ')            
            data.text = str(''.join(line.replace(" ",", ")))
        
        
    return root


'''
 witsml_fbe : Given the information, write the wistml.
'''
def witsml_fbe(datestring, xaxis, curves, curveData, data_style='UNKNOWN', depth_unit='m'):
    set_unknown_constants(data_style)
    #print('xaxis.shape',xaxis.shape)
    root = ET.Element(ROOT_TAG,ROOT_ATTR)
    wellset = ET.SubElement(root, WELLSET_TAG, WELLSET_ATTR)
    well = ET.SubElement(wellset, WELL_TAG, WELL_ATTR)
    

    wellname = ET.SubElement(well, WELL_NAME_TAG, WELL_NAME_ATTR)
    wellname.text = WELL_NAME_TEXT
    
    wellboreSet = ET.SubElement(well, WELLBORESET_TAG, WELLBORESET_ATTR)
    # The following level down is the wellbore
    wellbore = ET.SubElement(wellboreSet, WELLBORE_TAG, WELLBORE_ATTR)
    # wellbore has 4 elements
    wellborename = ET.SubElement(wellbore, WELLBORE_NAME_TAG, WELLBORE_NAME_ATTR)
    wellborename.text = WELLBORE_NAME_TEXT
    dtsInstalledSystemSet = ET.SubElement(wellbore, DTSINSTALLEDSYSTEMSET_TAG, DTSINSTALLEDSYSTEMSET_ATTR)
    dtsMeasurementSet = ET.SubElement(wellbore, DTSMEASUREMENTSET_TAG, DTSMEASUREMENTSET_ATTR)
    wellLogSet = ET.SubElement(wellbore, WELLLOGSET_TAG, WELLLOGSET_ATTR)
    # ....dtsInstalledSystemSet has information on the DTS/DAS system used
    dtsInstalledSystem = ET.SubElement(dtsInstalledSystemSet, DTSINSTALLEDSYSTEM_TAG, DTSINSTALLEDSYSTEM_ATTR)
    # ....|....dtsInstalledSystem has information on the DTS/DAS system used
    dtsInstalledSystemName = ET.SubElement(dtsInstalledSystem, DTSINSTALLEDSYSTEM_NAME_TAG, DTSINSTALLEDSYSTEM_NAME_ATTR)
    dtsInstalledSystemName.text = DTSINSTALLEDSYSTEM_NAME_TEXT
    fiberInformation = ET.SubElement(dtsInstalledSystem, FIBERINFORMATION_TAG, FIBERINFORMATION_ATTR)
    # ....|....|....fiberInformation - including mode
    fiber = ET.SubElement(fiberInformation, FIBER_TAG, FIBER_ATTR)
    # ....|....|....|....fiber - the name and mode of the fiber
    fiberName = ET.SubElement(fiber, FIBER_NAME_TAG, FIBER_NAME_ATTR)
    fiberName.text = FIBER_NAME_TEXT
    mode = ET.SubElement(fiber, FIBERMODE_TAG, FIBERMODE_ATTR)
    mode.text = FIBERMODE_TEXT
    # ....dtsMeasurementSet
    dtsMeasurement = ET.SubElement(dtsMeasurementSet, DTSMEASUREMENT_TAG, DTSMEASUREMENT_ATTR)
    # ....|....dtsMeasurement
    dtsMeasurementName = ET.SubElement(dtsMeasurement, DTSMEASUREMENT_NAME_TAG,DTSMEASUREMENT_NAME_ATTR)
    dtsMeasurementName.text = DTSMEASUREMENT_NAME_TEXT
    installedSystemUsed = ET.SubElement(dtsMeasurement, INSTALLEDSYSTEMUSED_TAG,INSTALLEDSYSTEMUSED_ATTR)
    installedSystemUsed.text = INSTALLEDSYSTEMUSED_TEXT
    dataInWellLog = ET.SubElement(dtsMeasurement, DATAINWELLLOG_TAG, DATAINWELLLOG_ATTR)
    dataInWellLog.text = WELLLOG_UNIQUE
    connectedToFiber = ET.SubElement(dtsMeasurement, CONNECTEDTOFIBER_TAG, CONNECTEDTOFIBER_ATTR)
    # ....wellLogSet
    wellLog = ET.SubElement(wellLogSet, WELLLOG_TAG, WELLLOG_ATTR)
    # ....|....wellLog
    wellLogName = ET.SubElement(wellLog, WELLLOG_NAME_TAG, WELLLOG_NAME_ATTR)
    wellLogName.text = WELLLOG_NAME_TEXT
    serviceCompany = ET.SubElement(wellLog, SERVICECOMPANY_TAG, SERVICECOMPANY_ATTR)
    serviceCompany.text = SERVICECOMPANY_NAME
    creationDate = ET.SubElement(wellLog, CREATIONDATE_TAG, CREATIONDATE_ATTR)
    #############LAS 2.0 INFO IN HERE.......###################
    creationDate.text = datestring
    indexType = ET.SubElement(wellLog, INDEXTYPE_TAG, INDEXTYPE_ATTR)
    indexType.text = INDEXTYPE_TEXT
    nullValue = ET.SubElement(wellLog, NULLVALUE_TAG, NULLVALUE_ATTR)
    nullValue.text = NULLVALUE_TEXT

    # The DEPTH part
    lognumber=1
    logCurveInfo = ET.SubElement(wellLog, LOGCURVEINFO_TAG, {'uid' : str(lognumber)} )
    blockInfo = ET.SubElement(wellLog, BLOCKINFO_TAG, {'uid' : str(lognumber)})
        
    mnemonic = ET.SubElement(logCurveInfo, MNEMONIC_TAG, MNEMONIC_ATTR)
    unit = ET.SubElement(logCurveInfo, UNIT_TAG, UNIT_ATTR)
    curveDescription = ET.SubElement(logCurveInfo,CURVEDESCRIPTION_TAG, CURVEDESCRIPTION_ATTR)
    mnemonic.text = 'DEPTH'
    unit.text = depth_unit
    curveDescription.text = 'depth along the fiber'
    indexType = ET.SubElement(blockInfo, INDEXTYPE_TAG, INDEXTYPE_ATTR)
    indexType.text = 'length'
    direction = ET.SubElement(blockInfo, DIRECTION_TAG, DIRECTION_ATTR)
    direction.text = 'increasing'
    indexCurve = ET.SubElement(blockInfo, INDEXCURVE_TAG, INDEXCURVE_ATTR)
    indexCurve.text = str(int(1))
    lognumber+=1

    # The data
    band=1
    # populate the column descriptions
    for curve in curves:
        logCurveInfo = ET.SubElement(wellLog, LOGCURVEINFO_TAG, {'uid' : str(lognumber)} )
        blockInfo = ET.SubElement(wellLog, BLOCKINFO_TAG, {'uid' : str(lognumber)})
        
        mnemonic = ET.SubElement(logCurveInfo, MNEMONIC_TAG, MNEMONIC_ATTR)
        unit = ET.SubElement(logCurveInfo, UNIT_TAG, UNIT_ATTR)
        curveDescription = ET.SubElement(logCurveInfo,CURVEDESCRIPTION_TAG, CURVEDESCRIPTION_ATTR)
        #'BAND-00'
        mnemonic.text = curve['mnemonic']
        #'unitless'
        unit.text = curve['unit']
        #'Frequency Band Energy 0-Nyquist'
        curveDescription.text = curve['description']

        blockCurveInfo = ET.SubElement(blockInfo, BLOCKCURVEINFO_TAG, BLOCKCURVEINFO_ATTR)
        curveId = ET.SubElement(blockCurveInfo, CURVEID_TAG, CURVEID_ATTR)
        curveId.text = str(int(band))
        columnIndex = ET.SubElement(blockCurveInfo, COLUMNINDEX_TAG, COLUMNINDEX_ATTR)
        columnIndex.text = curveId.text          
        band+=1
        lognumber+=1
    # populate the log data
    logData = ET.SubElement(wellLog, LOGDATA_TAG, LOGDATA_ATTR)

    xaxis = numpy.squeeze(xaxis)
    # sometimes we have only a single column...sometimes many columns...
    if curveData.ndim>1:
        for a in range(curveData.shape[0]):
            line = curveData[a]
            data = ET.SubElement(logData,DATA_TAG, DATA_ATTR)
            data.text = str(xaxis[a]) + ', ' + ', '.join(str(val) for val in line)
    else:
        for a in range(curveData.shape[0]):
            line = curveData[a]
            data = ET.SubElement(logData,DATA_TAG, DATA_ATTR)
            data.text = str(xaxis[a]) + ', ' +  str(line)        
    return root

'''
  writeFBE : Separates the writing part, in case we want to
             support things other than files
'''
def writeFBE(pathstem,fileout, root):
    tree = ET.ElementTree(root)
    ET.register_namespace("", "http://www.witsml.org/schemas/131")
    #fileout = filename.replace('.las','.fbe')
    xmlstr = minidom.parseString(ET.tostring(root)).toprettyxml(indent="   ")
    print(os.path.join(pathstem,fileout+'.fbe'))
    with open(os.path.join(pathstem,fileout+'.fbe'), "w") as f:
        f.write(xmlstr)


'''
 recursive_elem_counter : loops over XML recursively accumulating data locations
'''
def recursive_elem_counter(elem, list_of_files, depth_axis):
    for subelem in elem:
        if subelem.tag==LOGCURVEINFO_TAG:
            for subsubelem in subelem:
                if subsubelem.tag==MNEMONIC_TAG:
                    list_of_files.append(subsubelem.text)
        if subelem.tag==DATA_TAG:
            values = subelem.text.split(',')
            fvalues = [float(f) for f in values]
            depth_axis.append(fvalues[0])
        recursive_elem_counter(subelem, list_of_files, depth_axis)

'''
 recursive_elem_ingest : loops over XML recursively reading data
'''
def recursive_elem_ingest(elem, list_of_memmaps, idx, xidx, time_axis):
    for subelem in elem:
        if subelem.tag==DATA_TAG:
            values = subelem.text.split(',')
            fvalues = [float(f) for f in values]
            for a in range(len(list_of_memmaps)):
                list_of_memmaps[a][xidx,idx]=fvalues[a+1]
            xidx=xidx+1
        if subelem.tag==CREATIONDATE_TAG:
            value = datetime.datetime.strptime(subelem.text,"%Y-%m-%dT%H:%M:%S+00:00")
            time_axis.append(value.timestamp())
        recursive_elem_ingest(subelem, list_of_memmaps, idx, xidx, time_axis)   

'''
 recursive_elem_accumulator : used in the average and std. dev. determination
                              over multiple WITSML files
'''
def recursive_elem_accumulator(elem, values,squares, xidx):
    for subelem in elem:
        if subelem.tag==DATA_TAG:
            valueSet = subelem.text.split(',')
            fvalues = [float(f) for f in valueSet]
            for a in range(values.shape[1]):
                values[xidx,a]=values[xidx,a]+fvalues[a]
                squares[xidx,a]=squares[xidx,a]+(fvalues[a]*fvalues[a])
            xidx=xidx+1
        recursive_elem_accumulator(subelem, values, squares, xidx)   

'''
 recursive_elem_egest: used in writing WITSML to an XML structure for later
                       export
'''
def recursive_elem_egest(elem, values, xidx):
    for subelem in elem:
        if subelem.tag==DATA_TAG:
            subelem.text = ', '.join(str(val) for val in values[xidx])                
            xidx=xidx+1
        recursive_elem_egest(subelem, values, xidx)   

'''
 averageFBE :  Accumulate average FBE results over long time periods
               ...we take a whole directory in and write a whole directory out(?)
               ...with additional curves corresponding to std. dev.
'''
def averageFBE(datafiles, basedir, dirout, dirout_std):
    DEPTH = 'DEPTH'
    TIME = 'TIME'
    TMP = 'tmp'
    
    # The number of time samples
    nt = len(datafiles)

    
    # First file gives us nx and the list of output files
    filename = os.path.join(basedir,datafiles[0])
    tree = ET.parse(filename)
    root = tree.getroot()
    list_of_files = []
    depth_axis_values  =[]
    time_axis_values = []
    recursive_elem_counter(root, list_of_files, depth_axis_values)
    nx = len(depth_axis_values)

    set_of_values = numpy.zeros((nx,len(list_of_files)),dtype=numpy.double)
    set_of_squares = numpy.zeros((nx,len(list_of_files)),dtype=numpy.double)
    meanOut = numpy.zeros((nx,len(list_of_files)),dtype=numpy.double)
    stdOut = numpy.zeros((nx,len(list_of_files)),dtype=numpy.double)
        
    # loop and write...
    idx = 0
    for fname in datafiles:
        tree = ET.parse(os.path.join(basedir,fname))
        root = tree.getroot()
        xidx=0
        # to here
        recursive_elem_accumulator(root, set_of_values, set_of_squares, xidx)
        idx = idx + 1
        meanOut = set_of_values/idx
        stdOut = numpy.sqrt((set_of_squares/idx)-(meanOut*meanOut))
        # The stdOut first column should be MD just as for the meanOut
        stdOut[:,0]=meanOut[:,0]
        xidx=0
        recursive_elem_egest(root, meanOut, xidx)
        writeFBE(dirout,fname[:-4], root)
        # reset the root...
        xidx=0
        recursive_elem_egest(root, stdOut, xidx)
        writeFBE(dirout_std,fname[:-4], root)
'''
 readFBE : read a directory of *.fbe files and turn them into a directory containing a depth axis, a time axis and a 2D array of values
           - a format that is much easier to work with in reprocessing and interpretation flows.

def readFBE(datafiles,dirout):
    DEPTH = 'DEPTH'
    TIME = 'TIME'
    TMP = 'tmp'
    
    # The number of time samples
    nt = len(datafiles)

    
    # First file gives us nx and the list of output files
    filename = datafiles[0]
    tree = ET.parse(filename)
    root = tree.getroot()
    list_of_files = []
    depth_axis_values  =[]
    time_axis_values = []
    recursive_elem_counter(root, list_of_files, depth_axis_values)
    nx = len(depth_axis_values)

    # create memmap files - similar to the way we reprocess SGY to 1-second data chunks.
    memmapList = []
    tmpList = []
    tmpdir = os.path.join(dirout,TMP)
    if not os.path.exists(tmpdir):
        os.makedirs(tmpdir)
        
    for fname in list_of_files:
        if not fname==DEPTH:
            fileout = os.path.join(tmpdir, fname)
            memmapList.append(numpy.memmap(fileout, dtype=numpy.double, mode='w+', shape=(nx,nt)))
            tmpList.append(fileout)
    for mapped in memmapList:
        mapped.flush()
        
    # loop and write...
    idx = 0
    for fname in datafiles:
        tree = ET.parse(fname)
        root = tree.getroot()
        xidx=0
        recursive_elem_ingest(root, memmapList, idx, xidx, time_axis_values)
        idx = idx + 1

    # create time and depth axis as arrays
    time_axis = numpy.asarray(time_axis_values, dtype=numpy.double)
    depth_axis = numpy.asarray(depth_axis_values, dtype=numpy.double)

    for mapped in memmapList:
        mapped.flush()

    print("max values in memmaps")
    for a in range(len(memmapList)):
        print(numpy.max(memmapList[a]))
        
    print('Converting temporary files to permanent')
    for a in range(len(memmapList)):
        fileout = os.path.join(dirout,list_of_files[a+1]+'.npy')
        print(fileout)
        numpy.save(fileout, numpy.array(memmapList[a]))
        del mapped
        # override lazy behaviour...
        mapped = None
    memmapList = None
    # remove temporary files...
    for filename in tmpList:
        os.remove(filename)
    # Assumption that the fibre depth points do not change over the measurement period
    xaxisfilename = os.path.join(dirout,'measured_depth.npy')
    if not os.path.exists(xaxisfilename):
        numpy.save(xaxisfilename, depth_axis)
    taxisfilename = os.path.join(dirout,'time.npy')
    numpy.save(taxisfilename, time_axis)
'''       
'''
 readFBE : read a directory of *.fbe files and turn them into a directory containing a depth axis, a time axis and a 2D array of values
           - a format that is much easier to work with in reprocessing and interpretation flows.

           New readFBE to avoid dirswap and to make the results more compatible with distpy processing chains.
'''
def readFBE(datafiles,dirout,timestamp):
    DEPTH = 'DEPTH'
    TIME = 'TIME'
    TMP = 'tmp'
    
    # The number of time samples
    nt = len(datafiles)

    
    # First file gives us nx and the list of output files
    filename = datafiles[0]
    tree = ET.parse(filename)
    root = tree.getroot()
    list_of_files = []
    depth_axis_values  =[]
    time_axis_values = []
    recursive_elem_counter(root, list_of_files, depth_axis_values)
    nx = len(depth_axis_values)

    # create memmap files - similar to the way we reprocess SGY to 1-second data chunks.
    memmapList = []
    tmpList = []
    tmpdir = os.path.join(dirout,TMP)
    if not os.path.exists(tmpdir):
        os.makedirs(tmpdir)
        
    for fname in list_of_files:
        if not fname==DEPTH:
            filefolder = os.path.join(tmpdir,fname)
            if not os.path.exists(filefolder):
                os.makedirs(filefolder)
            fileout = os.path.join(tmpdir, fname, timestamp)
            memmapList.append(numpy.memmap(fileout, dtype=numpy.double, mode='w+', shape=(nx,nt)))
            tmpList.append(fileout)
    for mapped in memmapList:
        mapped.flush()
        
    # loop and write...
    idx = 0
    for fname in datafiles:
        tree = ET.parse(fname)
        root = tree.getroot()
        xidx=0
        recursive_elem_ingest(root, memmapList, idx, xidx, time_axis_values)
        idx = idx + 1

    # create time and depth axis as arrays
    time_axis = numpy.asarray(time_axis_values, dtype=numpy.double)
    depth_axis = numpy.asarray(depth_axis_values, dtype=numpy.double)

    for mapped in memmapList:
        mapped.flush()

    print("max values in memmaps")
    for a in range(len(memmapList)):
        print(numpy.max(memmapList[a]))
        
    print('Converting temporary files to permanent')
    for a in range(len(memmapList)):
        dir_final = os.path.join(dirout,list_of_files[a+1])
        if not os.path.exists(dir_final):
            os.makedirs(dir_final)
        fileout = os.path.join(dirout,list_of_files[a+1],timestamp+'.npy')
        print(fileout)
        numpy.save(fileout, numpy.array(memmapList[a]))
        del mapped
        # override lazy behaviour...
        mapped = None
    memmapList = None
    # remove temporary files...
    for filename in tmpList:
        os.remove(filename)
    # Assumption that the fibre depth points do not change over the measurement period
    for fname in list_of_files:
        if not fname==DEPTH:
            xaxisfilename = os.path.join(dirout,fname,'measured_depth.npy')
            if not os.path.exists(xaxisfilename):
                numpy.save(xaxisfilename, depth_axis)
            taxisfilename = os.path.join(dirout,fname,'time.npy')
            numpy.save(taxisfilename, time_axis)
'''
####################################################################################################
# Executable code...

#################################
# First make sure that the output
# directory has been created
#################################
if not os.path.exists(outpath):
    os.makedirs(outpath)

#################################
# Process all the files with the
# .las extension
#################################
onlyfiles = [f for f in listdir(inpath) if isfile(join(inpath, f))]
for filename in onlyfiles:
    if filename[-4:]=='.las':
        print('Converting ',filename)
        # we have a las file to convert
        root=las2fbe(inpath+filename)
        writeFBE(outpath,filename,root)
        
'''
def runtest():
    import os
    os.chdir("E:\\2019scr0001\\src2")
    dirout = "E:\\2019scr0001\\interpretation\\vlf"
    dirin = "E:\\2019scr0001\\results\\vlf"
    from witsml_fbe import readFBE
    readFBE(dirin,dirout)
    

if __name__ == "__main__":
    runtest()




# (C) 2020, Schlumberger. Refer to LICENSE

import json
import numpy
import h5py
import copy

#...pandas has good support for CSV files
import pandas
from datetime import datetime as dtime
try:
    from . import wistmlfbe
except:
    import distpy.io_help.witsmlfbe as witsmlfbe
import os
import itertools
import matplotlib.pyplot as plt
import matplotlib.colors
import matplotlib.dates as mdates
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

'''
 Automatic documentation support in latex and dot (graphviz)
 
'''
def substitute(lines, character, substitution):
    linesOut=[]
    for line in lines:
        linesOut.append(line.replace(character,substitution))
    return linesOut

def latexTop(lines):
    lines.append('\\documentclass{article}')
    # to allow us to incorporate figures...
    lines.append('\\usepackage{graphicx}')
    # to allow us to pretty-print JSON
    lines.append('\\usepackage{listings}')
    lines.append('\\usepackage{color}')
    lines.append('\\definecolor{dkgreen}{rgb}{0,0.6,0}')
    lines.append('\\definecolor{gray}{rgb}{0.5,0.5,0.5}')
    lines.append('\\definecolor{mauve}{rgb}{0.58,0,0.82}')

    lines.append('\\lstset{frame=tb,')
    lines.append('  language=Java,')
    lines.append('  aboveskip=3mm,')
    lines.append('  belowskip=3mm,')
    lines.append('  showstringspaces=false,')
    lines.append('  columns=flexible,')
    lines.append('  basicstyle={\\small\\ttfamily},')
    lines.append('  numbers=none,')
    lines.append('  numberstyle=\\tiny\\color{gray},')
    lines.append('  keywordstyle=\\color{blue},')
    lines.append('  commentstyle=\\color{mauve},')
    lines.append('  stringstyle=\\color{dkgreen},')
    lines.append('  breaklines=true,')
    lines.append('  breakatwhitespace=true,')
    lines.append('  tabsize=3')
    lines.append('}')
    lines.append('\\begin{document}')
    return lines
    
    
def latexTail(lines):
    lines.append('\\end{document}')
    return lines

def latexPng(lines):
    lines.append('\\subsection{Directed graph}')
    lines.append('\\begin{figure}')
    lines.append('\\includegraphics[width=\linewidth]{config.png}')
    lines.append('\\caption{The directed graph of the commands}')
    lines.append('\\label{fig:graph1}')
    lines.append('\\end{figure}')
    return lines


def latexJson(jsonIn,lines):
    lines.append('\\subsection{JSON  example}')
    lines.append('\\begin{lstlisting}')
    try:
        lines.append(json.dumps(jsonIn, indent=4, sort_keys=False))
    except TypeError:
        lines.append('Code not shown')
    lines.append('\\end{lstlisting}')
    return lines


def dot_graph(jsonArgs, command_list=None):
    lines = []
    fromList=[]
    nodeList=[]
    lines.append('digraph G {')
    if command_list is None:
        nodeList.append('load_data_0')
        # Basic graph
        for command in jsonArgs:
            comment = ''
            if 'comment' in command.keys():
                comment = command['comment']
            else:
                comment = command['name']
            formatting = command.get('formatting','')
            indx=command['uid']
            nodeList.append(command['name']+'_'+str(indx))
            lines.append(nodeList[indx] + '[label=\"'+comment+'\",'+formatting+'];')
            # from in_uid
            lines.append(nodeList[command['in_uid']]+' -> '+nodeList[indx] )
            prevlist = command.get('gather_uids',[-1])
            for prev in prevlist:
                if prev>=0:
                    lines.append(nodeList[prev]+' -> '+nodeList[indx])
    else:
        # We will pre-pend the list, so make sure we are working on a copy
        jsonArgs = copy.deepcopy(jsonArgs)
        jsonArgs.insert(0,{ "name" : "load_data", "uid" : 0, "in_uid" : -1 })
        # Advanced graphs
        for command, concreteCommand in zip(jsonArgs,command_list):
            indx=command['uid']
            localDoc = concreteCommand.docs()
            description = 'Undocumented feature'
            if 'one_liner' in localDoc:
                description = (localDoc['one_liner'])
            color = 'color=red,style=filled,fontcolor=white'
            if concreteCommand.isGPU()==True:
                color = 'color=green,style=filled,fontcolor=black'
            #label = '[label=\"'+command['name']+ ' '+description+'\",'+color+']'
            label = '[label=\"'+command['name']+ '\",'+color+']'
            nodeList.append(command['name']+'_'+str(indx))
            lines.append(nodeList[indx] + ' ' + label + ';')
            # from in_uid
            if command['in_uid']>=0:
                lines.append(nodeList[command['in_uid']]+' -> '+nodeList[indx] )
            prevlist = command.get('gather_uids',[-1])
            for prev in prevlist:
                if prev>=0:
                    lines.append(nodeList[prev]+' -> '+nodeList[indx])
    lines.append('}')
    return lines
    
#
# io_helpers.py
#     write2witsml   : output results to the real-time WITSML 1.3.1 format
#     thumbnail_plot : outputa 2D array (e.g. 1 second from SEGY) to a thumbnail image
#     json_io        : handle reading and writing of JSON to different storage targets
#     command2md     : markdown representation of a single command
def command2latex(command_list,linesOut,section='Command Set',topline='Commands used'):
    lines=[]
    lines.append('\\section{'+section+'}')
    lines.append(topline)
    for command in command_list:
        localDoc = command.docs()            
        lines.append('\\subsection{'+command.get_name()+'}')
        if 'one_liner' in localDoc:
            lines.append(localDoc['one_liner'])
        else:
            lines.append('Undocumented feature')
        if 'args' in localDoc:
            lines.append('\\begin{description}')
            for arg in localDoc['args']:
                lines.append('\\item['+arg+'] '+localDoc['args'][arg]['description'])
                if 'default' in localDoc['args'][arg]:
                    lines.append('\\\\default : '+str(localDoc['args'][arg]['default']))
            lines.append('\\end{description}')
        if command.isGPU()==True:
            lines.append('This command can be used with GPU.')
        else:
            lines.append('This command cannot be used with GPU.')
                 
    #doc.generate_pdf(fileout,clean_tex=False, compiler='C:\\Users\\miw\\AppData\\Local\\Programs\\MiKTeX 2.9\\miktex\\bin\\x64\\pdflatex.exe')
    lines = substitute(lines,'_','\\textunderscore ')
    for line in lines:
        linesOut.append(line)
    return linesOut


    
def command2md(command, lines):
    localDoc = command.docs()            
    lines.append('## '+command.get_name())
    if 'one_liner' in localDoc:
        lines.append(localDoc['one_liner'])
    if 'args' in localDoc:
        for arg in localDoc['args']:
            lines.append('`'+arg+'` : '+localDoc['args'][arg]['description'])
            if 'default' in localDoc['args'][arg]:
                lines.append('default : '+str(localDoc['args'][arg]['default']))
    if command.isGPU()==True:
        lines.append('This command can be used with GPU.')
    else:
        lines.append('This command cannot be used with GPU.')
    return lines

'''
 write2witsml : Export the results of stage 1 (strainrate-to-summary) as
                WITSML FBE format. This is compatible with Techlog.
'''
# Custom writer for FBE...
def write2witsml(dirout,fileout,datestring,xaxis, band00, data, low_freq, high_freq, prf, data_style='UNKNOWN', label_list=[], unit_list=[]):
    # handle the case of a transposed trace....
    if (data.shape[0]==1):
        data = numpy.transpose(data)
        band00 = numpy.transpose(band00)
        # simple axis replaces the actual axis - deferred better axis handling to a future version
        xaxis = numpy.linspace(0, data.shape[0], data.shape[0],endpoint=False,dtype=numpy.double)
    curves = witsmlfbe.generic_curves(low_freq,high_freq,prf)
    if (len(label_list)>0):
        icurve=0
        for label in label_list:
            curves[icurve]['mnemonic']=label
            curves[icurve]['description']=label
            icurve+=1
    if (len(unit_list)>0):
        icurve=0
        for unit in unit_list:
            curves[icurve]['unit']=unit
            icurve+=1

    dataOutsize = len(low_freq);
    useB0 = True
    if band00 is None:
        useB0 = False
    if useB0:
        dataOutsize=dataOutsize+1
    dataOut = numpy.zeros((data.shape[0],dataOutsize),dtype=numpy.double)
    if useB0:
        dataOut[:,0]=numpy.squeeze(band00)
        if (len(data.shape))<2:
            dataOut[:,1] = data
        else:
            dataOut[:,1:]=data
    else:
        dataOut = data
        curves = curves[1:]
    root = witsmlfbe.witsml_fbe(datestring,xaxis, curves, dataOut, data_style=data_style)
    # The datestring has characters that we do not want in a filename...
    # so we remo
    witsmlfbe.writeFBE(dirout,fileout,root)

'''
 csv2fbe - handles a basic CSV input, converting to WITSML FBE

 The CSV must be interpretable by pandas into a dataframe
'''
def csv2fbe(filein, stem, configJson):
    # Sometimes data comes in the DTS-style (depth as rows, time as columns), so we catch that and pass it on
    time_row = configJson.get('time_row',-1)
    if time_row>-1:
        return csv2dts(filein, stem, configJson)
    
    # output directory name
    head_tail = os.path.split(filein)
    # Assumption: original suffix is .csv or other 3 letter suffix.
    stage = (head_tail[1])[:-4]
    dirout = os.path.join(stem,stage)
    if not os.path.exists(dirout):
        os.makedirs(dirout)
    print('writing to '+dirout)

    # recover local variables from the config
    TIME_FORMAT = configJson.get('time_format','%Y-%m-%d %H:%M:%S')
    csv_labels = configJson.get('csv_labels',[])
    curves = configJson.get('curves',[])
    if not (len(csv_labels)==len(curves)):
        print('Different number of csv_labels to curves in the output file, please check the config.')
        return
    if (len(csv_labels)==0) or (len(curves)==0):
        print('Either no csv_labels or curves found, please check the config.')
        return
    depth_unit = configJson.get('depth_unit','m')
    data_style = configJson.get('data_style','UNKNOWN')

    # Read the CSV and sort by time-stamps
    df = pandas.read_csv(filein)
    uniquevals = numpy.unique(df[configJson['time_label']].values)
    
    
    for id in uniquevals:
        # extract all the data at the timestamp and sort by depth ready for output
        newdf = df[df[configJson['time_label']] == id]
        newdf = newdf.sort_values(by=[configJson['depth_label']])
        
        xaxis = numpy.array(newdf[configJson['depth_label']].tolist())
        
        timeval_curr = dtime.strptime(id,TIME_FORMAT)
        fileout = str(int(timeval_curr.timestamp()))
        datestring = dtime.fromtimestamp(int(fileout)).strftime("%Y-%m-%dT%H:%M:%S+00:00")

        # Convert the data to numpy for writing
        data = numpy.zeros((xaxis.shape[0],len(csv_labels)))
        for a in range(len(csv_labels)):
            data[:,a]=numpy.array(newdf[csv_labels[a]].tolist())
            
        root = witsmlfbe.witsml_fbe(datestring,xaxis, curves, data, data_style=data_style, depth_unit=depth_unit)
        # The datestring has characters that we do not want in a filename...
        # so we remo
        witsmlfbe.writeFBE(dirout,fileout,root)
'''
 csv2dts - handles gridded DTS data with rows of depth and
           columns that are timesteamps

 The CSV must be interpretable by pandas into a dataframe
'''
def csv2dts(filein, stem, configJson):
    # output directory name
    output_format = configJson.get('output_format','witsml')
    head_tail = os.path.split(filein)
    # Assumption: original suffix is .csv or other 3 letter suffix.
    stage = (head_tail[1])[:-4]
    dirout = os.path.join(stem,stage)
    if not os.path.exists(dirout):
        os.makedirs(dirout)
    print('writing to '+dirout)

    # recover local variables from the config
    TIME_FORMAT = configJson.get('time_format','%Y-%m-%d %H:%M:%S')
    # time_row must exist or we wouldn't assume this data format
    time_row = configJson['time_row']
    first_data = configJson.get('first_data',1)
    
    
    curves = configJson.get('curves',[ { "mnemonic" : "T", "unit" : "F", "description" : "DTS temperature in Fahrenheit" } ])
    depth_unit = configJson.get('depth_unit','m')
    data_style = configJson.get('data_style','UNKNOWN')

    skiprows = []
    for a in range(first_data):
        if not a==time_row:
            skiprows.append(a)

    # Read the CSV
    df = pandas.read_csv(filein,header=time_row,skiprows=skiprows)

    if output_format == 'witsml':
        firstTime=True
        for column in df:
            if firstTime == True:
                xaxis = numpy.array(df[column].tolist())
            else:
                timeval = dtime.strptime(column,TIME_FORMAT)
                fileout = str(int(timeval.timestamp()))
                datestring = dtime.fromtimestamp(int(fileout)).strftime("%Y-%m-%dT%H:%M:%S+00:00")
                data = numpy.array(df[column].tolist())
                root = witsmlfbe.witsml_fbe(datestring,xaxis, curves, data, data_style=data_style, depth_unit=depth_unit)
                # The datestring has characters that we do not want in a filename...
                # so we remo
                witsmlfbe.writeFBE(dirout,fileout,root)
            firstTime=False
    else:
        xaxis = numpy.squeeze(numpy.array(df.values[:,0]))
        cols = list(df.columns.values)
        taxis_dates = numpy.array(list(map(dtime.strptime, cols[1:], itertools.repeat(TIME_FORMAT,len(cols[1:])))))
        taxis = numpy.array([ts.timestamp() for ts in taxis_dates])
        data = df.values[:,1:]
        xaxisfilename = os.path.join(dirout,'measured_depth.npy')
        numpy.save(xaxisfilename, xaxis)
        taxisfilename = os.path.join(dirout,'time.npy')
        numpy.save(taxisfilename, taxis)
        datafilename = os.path.join(dirout,str(int(taxis[0]))+'.npy')
        numpy.save(datafilename, data)

'''
 Write a thumbnail plot using matplotlib
'''
def thumbnail_plot(stem_name, fname, data, xscale=[-1,-1],tscale=[-1,-1],plt_format='png',
                   std_val_mult=1.0, data2=None, jsonArgs={}):
    cmap=jsonArgs.get('cmap','viridis')
    axes=jsonArgs.get('axes',['depth (m)','time (s)'])
    xaxis=jsonArgs.get('xaxis',None)
    taxis=jsonArgs.get('taxis',None)
    invert_depth=jsonArgs.get('invert_depth',0)
    figure_size = jsonArgs.get('figure_size',(8,6))
    dots_per_inch = jsonArgs.get('dpi',300)
    plot_style = jsonArgs.get('plot_style','imshow')
    NLEVELS=30
    title_string = dtime.fromtimestamp(int(fname)).strftime('%Y-%m-%d %H:%M:%S')
    fig, ax = plt.subplots(figsize=figure_size, dpi=dots_per_inch)
    print('xscale',xscale)
    nx=data.shape[0]
    if xscale[0]==xscale[-1]:
        xscale=[0,nx]
    if data.ndim<2:
        plt.plot(xaxis,data)
        ax.set_xlim(xscale[0],xscale[-1])
        plt.title(title_string + ' (' + fname + ')', fontsize=10)
        return

    nt=data.shape[1]
    if tscale[0]==tscale[-1]:
        tscale=[0,nt]
        
    cscale=cmap
    #despiked = despike(data,3)
    meanval = numpy.mean(data)
    stdval = numpy.std(data)
    stdval = stdval*std_val_mult
    vminval = meanval - stdval
    vmaxval = meanval + stdval
    print(vminval,vmaxval)
    t_lims = tscale
    if tscale[0]>1e+7:
        # https://stackoverflow.com/questions/23139595/dates-in-the-xaxis-for-a-matplotlib-plot-with-imshow
        t_lims = list(map(dtime.fromtimestamp, [int(tscale[0]), int(tscale[1])]))
        t_lims = mdates.date2num(t_lims)

    taxis2= list(map(dtime.fromtimestamp, taxis.tolist()))
    taxis2= mdates.date2num(taxis2)

    if plot_style=='scatter':
        if data2 is None:
            plot_style='imshow'
        else:
            data_2 = data2[0]
            s=None
            c=None
            if len(data2)>1:
                c=data2[1].flatten()
            if len(data2)>2:
                s=data2[2].flatten()
            img = plt.scatter(data.flatten(),data_2.flatten(),c=c, s=s, cmap=cscale)
    if plot_style=='imshow':
        img = plt.imshow(data, cmap=cscale, aspect='auto', vmin=vminval, vmax=vmaxval, extent=[t_lims[0],t_lims[1],xscale[1],xscale[0]])
        cbar = fig.colorbar(img)
    if plot_style=='contourf':
        # create mesh to plot
        xv, yv = numpy.meshgrid(taxis2[:], xaxis[:])
        img = plt.contourf(xv,yv,data, NLEVELS,cmap=cscale)  # colormap,
        cbar = fig.colorbar(img)
    if tscale[0]>1e+7:
        # we have dates... so use dates
        # https://stackoverflow.com/questions/23139595/dates-in-the-xaxis-for-a-matplotlib-plot-with-imshow
        ax.set_xlim(t_lims[0],t_lims[1])
        ax.xaxis_date()
        date_format = mdates.DateFormatter('%Y-%m-%d %H:%M:%S')
        ax.xaxis.set_major_formatter(date_format)
        # set the dates to diagonal so they fit better
        fig.autofmt_xdate()
    plt.title(title_string + ' (' + fname + ')', fontsize=10)
    plt.xlabel(axes[1])
    plt.ylabel(axes[0])
    if invert_depth==1:
        plt.gca().invert_yaxis() # invert depth axis
    plt.rc('xtick',labelsize=15)
    plt.rc('ytick',labelsize=15)
    plt.tight_layout()

    
    fnameout = os.path.join(stem_name,fname+'.' + plt_format)
    print(fnameout)
    plt.savefig(os.path.join(stem_name,fname+'.'+plt_format),format=plt_format)
    inline_plots = jsonArgs.get('inline_plots',0)
    if inline_plots==1:
        plt.show()
    else:
        plt.close()


'''
 numpy_out - write the current data block to file.
'''
def numpy_out(stem_name,fname, data):
    numpy.save(os.path.join(stem_name,fname+'.npy'),data)
    
'''
 concatente_commands - take a command set and concatenate another command set.
                       No efficiency gains here...
'''
def concatenate_commands(orig_commands, new_commands):
    id_offset = orig_commands[-1]['uid']
    for command in new_commands:
        command['uid'] = command['uid']+id_offset
        command['in_uid'] = command['in_uid']+id_offset
        gather_uids = command.get('gather_uids',[])
        gather_uids_new = []
        for uid in gather_uids:
            gather_uids_new.append(uid+id_offset)
        if len(gather_uids_new)>0:
            command['gather_uids']=gather_uids_new
        orig_command.append(command)
    return orig_commands


'''
 write_plot : under development, this can replace plt.savefig() for compatibility with google cloud blob storage.
'''
def write_plot(stem_name, fname, plt):
    '''
    # For GOOGLE CLOUD FUNCTIONS USE...
    fig_to_upload = plt.gcf()
    # Save figure image to a bytes buffer
    buf = io.BytesIO()
    fig_to_upload.savefig(buf, format='png')
    buf.seek(0)
    # init GCS client and upload buffer contents
    blob = bucket.blob(OUT_PLOT_BLOB)  # This defines the path where the file will be stored in the bucket
    your_file_contents = blob.upload_from_string(buf.read(), content_type='image/png')
    '''
    # for everyone else...    
    fnameout = os.path.join(stem_name,fname+'.png')
    print(fnameout)
    plt.savefig(os.path.join(stem_name,fname+'.png'),format='png')
    plt.close()



'''
 json_io : all configuration files are JSON format. This class is already comfortable with both
           POSIX and blob-storage.
'''
# json_io allows the code to use local storage in provisioned mode...            
# NOTE: uses 'if' rather than Facade - currently there
#       are only minor differences between the codes
#       Later refactoring opportunity...
#       a facade would remove the 'if' options
class json_io(object):
    def __init__(self,filename, storage_flag):
        self.m_filename = filename
        self.m_provisioned = storage_flag
        #bucket = client.get_bucket(BACKSCATTER_BUCKET)
        self.m_json_blob = None
        self.m_json_file = None
        if self.m_provisioned==1:
            pass
        else:
            #self.m_json_blob = bucket.blob(filename)
            pass
    def write(self,string_file):
        if self.m_provisioned==1:
            with open(self.m_filename, 'w') as json_file:
                json_file.write(string_file)
        else:
            self.m_json_blob.upload_from_string(string_file)
    def read(self):
        if self.m_provisioned==1:
            with open(self.m_filename) as json_file:
                return json.load(json_file)
        else:
            configjson = self.m_json_blob.download_as_string()
            return json.loads(configjson.decode("utf-8"))

'''
 ingest_csv :  Reads a chunk from a CSV file - note that we have DTS data which
               is often double-ended
'''
def ingest_csv(filename, jsonArgs):
    TIME_FORMAT='%Y/%m/%d%H:%M:%S%p'

    # xdir is 1 for increasing and -1 for decreasing measured-depth
    time_start = jsonArgs.get('tmin',None)
    time_end = jsonArgs.get('tmax',None)
    
    # currently assume the CSV format corresponds to a typical DTS dataset
    # This means that the column headings are dates...
    topline = pandas.read_csv(filename,nrows=1)
    nt = topline.shape[1]-1
    print('nt=',nt)
    time_scale = numpy.zeros((nt),dtype=numpy.double)
    icol1 = -1
    icol2 = -1
    t1=-1
    t2=-1
    if time_start is not None:
        t1 = dtime.strptime(time_start.replace(" ",""),TIME_FORMAT).timestamp()
    if time_start is not None:
        t2 = dtime.strptime(time_end.replace(" ",""),TIME_FORMAT).timestamp()
    for a in range(nt):
        if a==0:
            print(topline.keys()[a+1])
        if a==nt-1:
            print(topline.keys()[a+1])
        # special case - don't know if this is standard date format for DTS
        timestring=topline.keys()[a+1]
        dt = dtime.strptime(timestring.replace(" ",""),TIME_FORMAT)    
        time_scale[a] = dt.timestamp()
        if a==0:
            print(time_scale[a])
        if a==nt-1:
            print(time_scale[a])
    print(t1,'-',time_scale[0],'=',t1-time_scale[0])
    print(time_scale[-1],'-',t2,'=',time_scale[-1]-t2)
    if time_start is None:
        icol1=0
    else:                
        icol1 = numpy.argmin(numpy.abs(time_scale-t1))
        if icol1>nt-2:
            print('Time out of range')
            icol1=nt-2
    if time_end is None:
        icol2 = nt
    else:
        icol2 = numpy.argmin(numpy.abs(time_scale-t2))
        if icol2<=icol1:
            print('Single time value')
            icol2=icol1+1
    time_scale=time_scale[icol1:icol2]
    print(t1,t2,icol1,icol2)
    print(filename)
    print(topline.shape)
    topline=pandas.read_csv(filename,skiprows=10,usecols=[0])
    print('one column ',topline.shape)
    measured_depth=topline.astype(numpy.double)
    topline=pandas.read_csv(filename,skiprows=10,usecols=range(icol1+1,icol2+1))
    print('data ',topline.shape)
    data = topline.astype(numpy.double)
    
    jsonArgs['xaxis']=measured_depth
    jsonArgs['taxis']=time_scale
    return data, time_scale, jsonArgs

'''
 systemConfig : reads and interprets the JSON file for system configuration
'''
def systemConfig(configOuterFile):
    # an outer configuration file...
    configOuterJson = json_io(configOuterFile,1)
    csData = configOuterJson.read()
    inpath = csData.get('in_directory', None)
    outpath =csData.get('out_directory', None)
    #...better option
    indrive = csData.get('in_drive',None)
    if not (indrive is None):
        inpath = os.path.join(csData['in_drive'],csData['project'],csData['data'])
        outpath  = os.path.join(csData['out_drive'],csData['project'],csData['results'])
    xaxisfile = os.path.join(inpath,csData.get('measured_depth','none.npy'))
    taxisfile = os.path.join(inpath,csData.get('time_scale','none.npy'))
    prf = csData.get('prf',10000)

    BOX_SIZE = csData.get('BOX_SIZE',400)
    
    configFile = csData.get('config', None)
    parallel_flag = csData.get('parallel_flag',0)
    PARALLEL = False
    NCPU=1
    if parallel_flag>0:
        PARALLEL = True
        NCPU = csData.get('ncpu',1)
    return inpath, outpath, configFile, PARALLEL, NCPU, BOX_SIZE, xaxisfile, taxisfile, prf



# (C) 2020, Schlumberger. Refer to package LICENSE.

'''
 CommandFactory - governs the use of commands in distpy
                  for processing and plotting
# There is a special command at the head-node - where data is loaded
'''
import numpy
import types
import distpy.calc.extra_numpy as extra_numpy
import distpy.calc.pub_command_set as pub_command_set
# - example of extension
# import distpy.calc.my_command_set as my_command_set
import distpy.calc.plt_command_set as plt_command_set


def CommandFactory(commandList, commandJson, extended_list=[]):
    # currently we are just extending our own list...
    knownList = {}
    # Add your own command sets below. The order indicates
    # precidence. For example my_command_set is overwriting
    # commands  in the public command set.
    knownList = pub_command_set.KnownCommands(knownList)
    # plotting subset
    knownList = plt_command_set.KnownCommands(knownList)
    # - load any supplied extensions
    for module_name in extended_list:
        exec('import '+module_name)
        exec('knownList = '+module_name+'.KnownCommands(knownList)')
    
    name = commandJson.get('name','NONE')
    plot_type = commandJson.get('plot_type','NONE')
    #print(name)
    # multiple entries for the previous command
    # to support the n-to-1 paths
    previous = commandJson.get('in_uid',-1)
    prevlist = commandJson.get('gather_uids',[-1])
    prev_stack = []
    for prev in prevlist:
        if prev>=0:
            prev_stack.append(commandList[prev])
    commandJson['commands']=prev_stack
    if previous>=0:
        prev_result = commandList[previous]
        if name not in knownList:
            name = 'NONE'
        return (knownList[name](prev_result,commandJson))
        
    return None



# (C) 2020, Schlumberger. Refer to LICENSE

########################################
# USER CONFIGURABLE VALUES:
basedir = "{PATH}\\witsml"
dirout = "{PATH}\\results"
PARALLEL=True
NCPU=1
repackage_size=10
########################################
import argparse

import os
import distpy.io_help.witsmlfbe as witsmlfbe
import distpy.io_help.io_helpers as io_helpers
import distpy.io_help.directory_services as ds
import multiprocessing

'''
 readFBEController : dependencies are captured here to keep the main() as
                     generic as possible. In the future refactoring might
                     result in one parallel controller.
'''
def readFBEController(dirin,dirout, repackage_size):
    datafiles=[]
    prefixes = []
    for root, dirs, files in os.walk(dirin):
        for datafile in files:
            if datafile[-3:]=='fbe':
                prefixes.append(datafile[:-4])
                datafiles.append(os.path.join(root,datafile))
    # alphanumeric sorting is assumed to work...
    # a more detail-oriented approach would extract all the unixtime stamps in key-value pairs and sort on those.
    datafiles.sort()
    nt=len(datafiles)
    if (repackage_size<1):
        repackage_size=nt
    # repackage size is now only zero if there are no fbe files found..
    if (repackage_size>1):
        for a in range(0,nt,repackage_size):
            endIdx = a+repackage_size
            if endIdx>=nt:
                endIdx=nt-1
            datafileset = datafiles[a:endIdx]
            witsmlfbe.readFBE(datafileset,dirout,prefixes[a])
        
def main(configOuterFile):
    # standard configuration information
    basedir,dirout,configFile,PARALLEL,NCPU,repackage_size, xaxisfile,taxisfile,prf = io_helpers.systemConfig(configOuterFile)

    #scan for directories
    witsml_dirs=[]
    results_dirs=[]
    for root, dirs, files in os.walk(basedir):
        for resultdir in dirs:
            print(resultdir)
            # NOTE we handle the 1 exception directly.
            # in the future something more general may be needed
            if not resultdir=='png':
                witsml_dirs.append(os.path.join(root,resultdir))
                onedirout = os.path.join(dirout,resultdir)
                results_dirs.append(onedirout)
                ds.makedir(onedirout)

    # packaging the WITSML into chunks - in case there were days and days of DAS recording so that
    # the data are still too big to handle...
    if not PARALLEL:
        # parallel does not work in Techlog...
        for (witsmldir,resultdir) in zip(witsml_dirs,results_dirs):
            print(resultdir)            
            readFBEController(witsmldir,resultdir,repackage_size)
    else:
        manager = multiprocessing.Manager()
        q = manager.Queue()
        print(multiprocessing.cpu_count())
        
        pool = multiprocessing.Pool(processes=NCPU, maxtasksperchild=1)
        
        jobs = []
        for (witsmldir,resultdir) in zip(witsml_dirs,results_dirs):
            job = pool.apply_async(readFBEController, [witsmldir,resultdir,repackage_size])
            print(job)
            jobs.append(job)

        for job in jobs:
            job.get()
            print(job)

        q.put('kill')
        pool.close()

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-f","--filename", help='A configuration file')
    args = parser.parse_args()
    main(args.filename)




from __future__ import division, absolute_import, print_function

__all__ = ['cupy_facade','agnostic','extra_numpy','extra_pyplot','unit_handler',
           'pub_command_set','plt_command_set','processing_commands.py']


# (C) 2020, Schlumberger. Refer to LICENSE



# (C) 2020 Schlumberger. Refer to LICENSE

'''
    Extra functions needed for matplotlib.pyplot output

    NOTE: 2D signal processing belongs in extra_numpy
          only actual matplotlib output stuff should be in here
          
extra_pyplot
'''
import matplotlib.pyplot as plt
import matplotlib.colors
import matplotlib.dates as mdates
import matplotlib.patches as patches
import time
import numpy

'''
 Color-scales - in general prefer the python visually uniform color scales
                since there is much less chance of misinterpretation

 The following approximation to the UAVK color scale from Techlog is made
 available here. Use with caution!
'''
def UAVK():
    UAVK_list = ["white","navy","blue","royalblue","deepskyblue","cyan",
                 "palegreen","greenyellow","yellow","orange",
                 "darkorange","orangered","tomato","red","firebrick",
                 "brown"]
    return UAVK_list

'''
 All the python colorscales such as magnma and viridis, plus the UAVK
'''
def supported_colorscales(request):
    cscale=request
    if (request == 'UAVK'):
        cscale = matplotlib.colors.LinearSegmentedColormap.from_list("", UAVK())
    return cscale
    
'''
 add_user_events : labelled time-stamps to annotate plots
'''
def add_user_events(event_list,plt, time_ref):
    # Add user events
    index1 = 0
    for label in event_list:
        event_start = label['event_label_start']
        event_mark  = label['event_mark']
        event_text  = label['event_text']
        
        event_time=time.mktime((time.strptime(event_mark,'%Y-%m-%d %H:%M:%S')))
        event_line = numpy.searchsorted(time_ref, event_time)
        plt.axvline(event_line,color='r')
        plt.text(event_line,event_start,event_text,rotation='vertical',fontsize=7, bbox=dict(facecolor='yellow', alpha=0.85))
        index1 += 1

'''
 add_user_zones : labelled depth extents to annotate plots, for example
                  with perforation cluster locations
'''
def add_user_zones(event_list, plt, depth_ref, line_color):
    # Add user events
    index1 = 0
    for label in event_list:
        start = label['start']
        end   = label['end']
        name  = label['name']
        
        depth_line = numpy.searchsorted(depth_ref, start)
        plt.axhline(depth_line,color=line_color)
        depth_line = numpy.searchsorted(depth_ref, end)
        plt.axhline(depth_line,color=line_color)
        # patches.Rectangle((location), width, height, ...)
        #rect = patches.Rectangle((50,100),40,30,linewidth=1,edgecolor='r',facecolor='none')
        plt.text(0,depth_line,name,rotation='horizontal',fontsize=7, bbox=dict(facecolor='yellow', alpha=0.85))

'''
 time_stamps : turn a list of unixtimestamps into a lits of string-based
               timestamps for writing on plots
'''
def time_stamps(time_ref):
    time_list = []
    for seconds in time_ref:
        time_list.append(time.asctime(time.gmtime(seconds)))
    return time_list

'''
 add_fibre_ends : draw lines corresponding to the start and end of the fibre
                  Typically DAS data will have data before the wellhead, and
                  noise corresponding to backscatter that occurred beyond the
                  two-way travel time.
'''
def add_fibre_ends(start_of_fibre,end_of_fibre,plt):
    # Start and end of fibre
    plt.axhline(start_of_fibre,color='w')
    plt.axhline(end_of_fibre,color='w')




# (C) 2020, Schlumberger. Refer to LICENSE

########################################
# USER CONFIGURABLE VALUES:
basedir = "{PATH}\\RawData"
dirout = "{PATH}\\backscatter"
PARALLEL=True
NCPU=1
########################################
import argparse

import os
import numpy
import distpy.io_help.io_helpers as io_helpers
import distpy.io_help.directory_services as ds
import multiprocessing

'''
 ingets1csv : dependencies are captured in this function rather than main()
              This keeps the main() as generic as possible so that a future
              refactoring might be able to standardize the parallel run controller.
'''
def ingest1csv(fname, dirout, jsonArgs):
    data, timestamp, jsonArgs = io_helpers.ingest_csv(fname,jsonArgs)
    fname=str(int(timestamp[0]))+'.npy'
    numpy.save(os.path.join(dirout,fname),data)
    numpy.save(os.path.join(dirout,'measured_depth.npy'),jsonArgs['xaxis'])
    numpy.save(os.path.join(dirout,'time.npy'),jsonArgs['taxis'])
    


def main(configOuterFile):
    # standard configuration information
    basedir,dirout,jsonConfig,PARALLEL,NCPU,BOX_SIZE, xaxisfile,taxisfile,prf = io_helpers.systemConfig(configOuterFile)
    ds.makedir(dirout)

    configData = {}
    if jsonConfig is not None:
        configFile = io_helpers.json_io(jsonConfig,1)
        configData = configFile.read()
    #scan for directories
    csv_files=[]
    for root, dirs, files in os.walk(basedir):
        for datafile in files:
            csv_files.append(os.path.join(root,datafile))

    if not PARALLEL:
        # parallel does not work in Techlog...
        for fname in csv_files:
            print(fname)
            io_helpers.csv2fbe(fname,dirout, configData)
    else:
        manager = multiprocessing.Manager()
        q = manager.Queue()
        print(multiprocessing.cpu_count())

        pool = multiprocessing.Pool(processes=NCPU, maxtasksperchild=1)

        jobs = []
        for fname in csv_files:
            job = pool.apply_async(io_helpers.csv2fbe,[fname,dirout, configData])
            print(job)
            jobs.append(job)

        for job in jobs:
            job.get()
            print(job)

        q.put('kill')
        pool.close()

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-f","--filename", help='A configuration file')
    args = parser.parse_args()
    main(args.filename)



# (c) 2020, Schlumberger. Refer to LICENSE

import os
import numpy

dirintest = "{PATH}\\results\\vlf"

'''
  timeshift : A utility for adding or subtracting a number of hours on
              unixtimestamp filenames.
              
'''
def timeshift(dirin, hours=0.0):
    MD_NAME = 'measured_depth.npy'
    TIME_NAME = 'time.npy'
    seconds = int(hours*3600)
    if seconds==0:
        return
    
    dirnames = [d for d in os.listdir(dirin) if os.path.isdir(os.path.join(dirin,d))]
    testdir = os.path.join(dirin,dirnames[0])
    fnames = [f for f in os.listdir(testdir) if os.path.isfile(os.path.join(testdir,f))]
    fnames.sort()
    if seconds>0:
        fnames.reverse()
    # make directories for the results
    xaxis=None
    for filename in fnames:
        if filename[-3:]=='npy':
            if filename==MD_NAME:
                xaxis=numpy.load(os.path.join(testdir,filename))
            dirname = filename[:-4]
            unixtime = int(dirname)+seconds
            
            fulldirin = os.path.join(dirin,filename)
            fulldirout = os.path.join(dirin,str(unixtime)+'.npy')
            os.rename(fulldirin,fulldirout)

# testing
if __name__ == "__main__":
    timeshift(dirintest)
